"""AWS Secrets Manager adapter."""

from __future__ import annotations

import asyncio
import time
from collections.abc import Awaitable, Callable
from typing import Any

from pydantic import BaseModel, Field

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class AWSSecretManagerSettings(BaseModel):
    """Settings for interacting with AWS Secrets Manager."""

    region: str = Field(description="AWS region where the secrets live.")
    endpoint_url: str | None = Field(
        default=None, description="Optional custom endpoint (e.g., LocalStack)."
    )
    profile_name: str | None = Field(
        default=None, description="AWS profile name for credential discovery."
    )
    access_key_id: str | None = Field(default=None)
    secret_access_key: str | None = Field(default=None)
    session_token: str | None = Field(default=None)
    cache_ttl_seconds: int = Field(default=60, ge=0)
    healthcheck_secret: str | None = Field(
        default=None, description="Optional secret name probed during health checks."
    )
    version_stage: str | None = Field(
        default=None, description="Optional version stage filter (e.g., AWSCURRENT)."
    )


class AWSSecretManagerAdapter:
    """Async adapter powered by aioboto3's Secrets Manager client."""

    metadata = AdapterMetadata(
        category="secrets",
        provider="aws-secret-manager",
        factory="oneiric.adapters.secrets.aws:AWSSecretManagerAdapter",
        capabilities=["remote", "cache"],
        stack_level=52,
        priority=560,
        source=CandidateSource.LOCAL_PKG,
        owner="Security",
        requires_secrets=True,
        settings_model=AWSSecretManagerSettings,
    )

    def __init__(
        self,
        settings: AWSSecretManagerSettings,
        *,
        client: Any | None = None,
        client_factory: Callable[[], Awaitable[Any]] | None = None,
    ) -> None:
        self._settings = settings
        self._client = client
        self._client_factory = client_factory
        self._client_cm: Any | None = None
        self._cache: dict[str, tuple[str, float]] = {}
        self._cache_lock = asyncio.Lock()
        self._logger = get_logger("adapter.secrets.aws").bind(
            domain="adapter",
            key="secrets",
            provider="aws-secret-manager",
            region=settings.region,
        )

    async def init(self) -> None:
        if self._client is not None:
            return
        if self._client_factory:
            self._client = await self._client_factory()
            return
        try:
            import aioboto3
        except ModuleNotFoundError as exc:  # pragma: no cover - optional dependency
            raise LifecycleError("aioboto3-missing") from exc

        session_kwargs: dict[str, Any] = {
            "region_name": self._settings.region,
        }
        if self._settings.profile_name:
            session_kwargs["profile_name"] = self._settings.profile_name
        session = aioboto3.Session(**session_kwargs)
        client_kwargs: dict[str, Any] = {
            "service_name": "secretsmanager",
            "endpoint_url": self._settings.endpoint_url,
            "aws_access_key_id": self._settings.access_key_id,
            "aws_secret_access_key": self._settings.secret_access_key,
            "aws_session_token": self._settings.session_token,
        }
        client_kwargs = {k: v for k, v in client_kwargs.items() if v is not None}
        self._client_cm = session.client(**client_kwargs)
        self._client = await self._client_cm.__aenter__()
        self._logger.info("adapter-init", adapter="aws-secret-manager")

    async def health(self) -> bool:
        probe = self._settings.healthcheck_secret
        if not probe:
            return True
        try:
            value = await self.get_secret(probe, allow_missing=True)
            return value is not None
        except Exception as exc:  # pragma: no cover - external failure path
            self._logger.warning("adapter-health-error", error=str(exc))
            return False

    async def cleanup(self) -> None:
        if self._client_cm:
            await self._client_cm.__aexit__(None, None, None)
        if self._client and hasattr(self._client, "close"):
            await self._client.close()
        self._client = None
        self._client_cm = None
        self._cache.clear()
        self._logger.info("adapter-cleanup-complete", adapter="aws-secret-manager")

    async def invalidate_cache(self) -> None:
        async with self._cache_lock:
            self._cache.clear()
        self._logger.info("secrets-cache-invalidated", adapter="aws-secret-manager")

    async def get_secret(
        self,
        name: str,
        *,
        version_stage: str | None = None,
        allow_missing: bool = False,
    ) -> str | None:
        cached = await self._get_cached(name, version_stage)
        if cached is not None:
            return cached
        client = self._ensure_client()
        request: dict[str, Any] = {"SecretId": name}
        stage = version_stage or self._settings.version_stage
        if stage:
            request["VersionStage"] = stage
        try:
            response = await client.get_secret_value(**request)
        except Exception as exc:
            if allow_missing and self._is_not_found(exc):
                return None
            raise
        value = self._extract_secret(response)
        await self._set_cached(name, version_stage, value)
        return value

    def _ensure_client(self) -> Any:
        if not self._client:
            raise LifecycleError("aws-secret-client-not-initialized")
        return self._client

    def _extract_secret(self, response: dict[str, Any]) -> str:
        if "SecretString" in response:
            return str(response["SecretString"])
        if "SecretBinary" in response:
            data = response["SecretBinary"]
            if isinstance(data, (bytes, bytearray)):
                return data.decode("utf-8")
        raise LifecycleError("aws-secret-invalid-response")

    def _is_not_found(self, exc: Exception) -> bool:
        code = getattr(exc, "response", {}).get("Error", {}).get("Code")
        if isinstance(code, str) and code in {"ResourceNotFoundException", "404"}:
            return True
        message = getattr(exc, "args", [None])[0]
        if isinstance(message, str):
            return "ResourceNotFound" in message or "404" in message
        return False

    async def _get_cached(self, name: str, version_stage: str | None) -> str | None:
        async with self._cache_lock:
            key = self._cache_key(name, version_stage)
            cached = self._cache.get(key)
            if not cached:
                return None
            value, expires_at = cached
            if expires_at < time.monotonic():
                self._cache.pop(key, None)
                return None
            return value

    async def _set_cached(
        self, name: str, version_stage: str | None, value: str
    ) -> None:
        if self._settings.cache_ttl_seconds == 0:
            return
        async with self._cache_lock:
            key = self._cache_key(name, version_stage)
            expires_at = time.monotonic() + self._settings.cache_ttl_seconds
            self._cache[key] = (value, expires_at)

    def _cache_key(self, name: str, version_stage: str | None) -> str:
        return f"{name}:{version_stage or self._settings.version_stage or 'default'}"



"""Security validation helpers for Oneiric.

This module provides security controls for factory resolution, input validation,
and other security-critical operations.
"""

from __future__ import annotations

import os
import re
from typing import Any

# Factory string format: module.path:function_name
FACTORY_PATTERN = re.compile(
    r"^[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*:[a-zA-Z_][a-zA-Z0-9_]*$"
)

# Default safe prefixes (allow only oneiric and explicitly approved packages)
DEFAULT_ALLOWED_PREFIXES = [
    "oneiric.",
    # Users can extend via ONEIRIC_FACTORY_ALLOWLIST environment variable
]

# Modules that are always blocked (security risk)
BLOCKED_MODULES = [
    "os",
    "subprocess",
    "sys",
    "importlib",
    "__builtin__",
    "builtins",
    "shutil",
    "pathlib",  # Can be used for path manipulation
    "tempfile",  # Can write to filesystem
]


def validate_factory_string(
    factory: str,
    allowed_prefixes: list[str] | None = None,
) -> tuple[bool, str | None]:
    """Validate factory string format and module path.

    Args:
        factory: Factory string in format "module.path:function"
        allowed_prefixes: List of allowed module prefixes (defaults to DEFAULT_ALLOWED_PREFIXES)

    Returns:
        Tuple of (is_valid, error_message). error_message is None if valid.

    Examples:
        >>> validate_factory_string("oneiric.demo:DemoAdapter")
        (True, None)

        >>> validate_factory_string("os:system")
        (False, "Factory module 'os' is blocked for security reasons")

        >>> validate_factory_string("invalid")
        (False, "Invalid factory format: invalid. Expected 'module.path:function'")
    """
    # Format validation
    if not FACTORY_PATTERN.match(factory):
        return (
            False,
            f"Invalid factory format: {factory}. Expected 'module.path:function'",
        )

    module_path, _, attr = factory.partition(":")

    # Check against blocked modules first (highest priority)
    for blocked in BLOCKED_MODULES:
        if module_path == blocked or module_path.startswith(f"{blocked}."):
            return (
                False,
                f"Factory module '{module_path}' is blocked for security reasons",
            )

    # Check against allowlist
    prefixes = (
        allowed_prefixes if allowed_prefixes is not None else DEFAULT_ALLOWED_PREFIXES
    )

    # If allowlist is empty, reject all
    if not prefixes:
        return (
            False,
            f"Factory module '{module_path}' not in allowlist (allowlist is empty)",
        )

    if not any(module_path.startswith(prefix) for prefix in prefixes):
        return (
            False,
            f"Factory module '{module_path}' not in allowlist. Allowed prefixes: {prefixes}",
        )

    return True, None


def load_factory_allowlist() -> list[str]:
    """Load factory allowlist from environment or config.

    Reads from ONEIRIC_FACTORY_ALLOWLIST environment variable.
    Format: comma-separated list of module prefixes.

    Returns:
        List of allowed module prefixes

    Examples:
        # With env: ONEIRIC_FACTORY_ALLOWLIST="oneiric.,myapp.,vendor."
        >>> load_factory_allowlist()
        ['oneiric.', 'myapp.', 'vendor.']

        # Without env (defaults)
        >>> load_factory_allowlist()
        ['oneiric.']
    """
    env_value = os.getenv("ONEIRIC_FACTORY_ALLOWLIST")
    if env_value is not None:
        # Environment variable is set (even if empty)
        # Parse comma-separated list and ensure trailing dots
        prefixes = []
        for prefix in env_value.split(","):
            prefix = prefix.strip()
            if prefix and not prefix.endswith("."):
                prefix += "."
            if prefix:
                prefixes.append(prefix)
        return prefixes
    return DEFAULT_ALLOWED_PREFIXES.copy()


def validate_key_format(key: str, allow_dots: bool = True) -> tuple[bool, str | None]:
    """Validate component key format (prevent path traversal).

    Args:
        key: Component key to validate
        allow_dots: Whether to allow dots in keys (default: True)

    Returns:
        Tuple of (is_valid, error_message)

    Examples:
        >>> validate_key_format("my-component")
        (True, None)

        >>> validate_key_format("../../evil")
        (False, "Key contains path traversal: ../../evil")
    """
    if not key:
        return False, "Key cannot be empty"

    # Check for path traversal attempts
    if ".." in key or key.startswith("/") or "\\" in key:
        return False, f"Key contains path traversal: {key}"

    # Validate character set (alphanumeric, dash, underscore, optionally dot)
    if allow_dots:
        pattern = r"^[a-zA-Z0-9_\-\.]+$"
    else:
        pattern = r"^[a-zA-Z0-9_\-]+$"

    if not re.match(pattern, key):
        allowed = "alphanumeric with -_." if allow_dots else "alphanumeric with -_"
        return False, f"Key contains invalid characters (must be {allowed}): {key}"

    return True, None


def validate_priority_bounds(priority: Any) -> tuple[bool, str | None]:
    """Validate priority is within acceptable bounds.

    Args:
        priority: Priority value to validate

    Returns:
        Tuple of (is_valid, error_message)
    """
    MIN_PRIORITY = -1000
    MAX_PRIORITY = 1000

    if not isinstance(priority, int):
        return False, f"Priority must be integer, got {type(priority).__name__}"

    # Type guard satisfied - priority is int
    if priority < MIN_PRIORITY or priority > MAX_PRIORITY:
        return (
            False,
            f"Priority {priority} out of bounds [{MIN_PRIORITY}, {MAX_PRIORITY}]",
        )

    return True, None


def validate_stack_level_bounds(stack_level: Any) -> tuple[bool, str | None]:
    """Validate stack_level is within acceptable bounds.

    Args:
        stack_level: Stack level value to validate

    Returns:
        Tuple of (is_valid, error_message)
    """
    MIN_STACK_LEVEL = -100
    MAX_STACK_LEVEL = 100

    if not isinstance(stack_level, int):
        return False, f"Stack level must be integer, got {type(stack_level).__name__}"

    # Type guard satisfied - stack_level is int
    if stack_level < MIN_STACK_LEVEL or stack_level > MAX_STACK_LEVEL:
        return (
            False,
            f"Stack level {stack_level} out of bounds [{MIN_STACK_LEVEL}, {MAX_STACK_LEVEL}]",
        )

    return True, None



"""Entry-point discovery helpers for Oneiric plugins."""

from __future__ import annotations

from collections import abc
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass, field
from importlib import metadata

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.config import PluginsConfig
from oneiric.core.logging import get_logger
from oneiric.core.resolution import Candidate, Resolver

logger = get_logger("plugins")

DEFAULT_ENTRY_POINT_GROUPS = (
    "oneiric.adapters",
    "oneiric.services",
    "oneiric.tasks",
    "oneiric.events",
    "oneiric.workflows",
)


@dataclass
class PluginEntryRecord:
    group: str
    entry_point: str
    payload_type: str
    registered_candidates: int


@dataclass
class PluginErrorRecord:
    group: str
    entry_point: str
    reason: str


@dataclass
class PluginRegistrationReport:
    """Diagnostics describing plugin bootstrap activity."""

    groups: list[str] = field(default_factory=list)
    registered: int = 0
    entries: list[PluginEntryRecord] = field(default_factory=list)
    errors: list[PluginErrorRecord] = field(default_factory=list)

    @classmethod
    def empty(cls) -> PluginRegistrationReport:
        return cls()

    def as_dict(self) -> dict:
        return {
            "groups": self.groups.copy(),
            "registered": self.registered,
            "entries": [entry.__dict__ for entry in self.entries],
            "errors": [error.__dict__ for error in self.errors],
        }


@dataclass
class _FactoryLoadResult:
    group: str
    entry_point: str
    factory: Callable[[], object] | None = None
    error: str | None = None


def iter_entry_points(group: str) -> Sequence[metadata.EntryPoint]:
    """Return entry points for the supplied group (PyPI metadata helper)."""

    entry_points = metadata.entry_points()
    if hasattr(entry_points, "select"):
        return tuple(entry_points.select(group=group))
    # Fallback for dict-like EntryPoints on older Python
    if hasattr(entry_points, "get"):
        return tuple(entry_points.get(group, []))  # type: ignore[attr-defined]
    return ()


def load_callables(group: str) -> list[Callable[[], object]]:
    """Load callable factories exposed via an entry-point group.

    Any entry point that fails to load is logged and skipped so plugin discovery
    remains best-effort.
    """

    return [
        result.factory
        for result in _load_entry_point_factories(group)
        if result.factory
    ]


def discover_metadata(group: str) -> Iterable[object]:
    """Convenience wrapper that calls each plugin factory and yields metadata."""

    for factory in load_callables(group):
        try:
            yield factory()
        except Exception as exc:  # pragma: no cover - logging guard
            logger.warning(
                "plugin-metadata-failed",
                group=group,
                factory=getattr(factory, "__name__", repr(factory)),
                error=str(exc),
            )


def register_entrypoint_plugins(
    resolver: Resolver,
    config: PluginsConfig,
    *,
    skip_if_loaded: bool = True,
) -> PluginRegistrationReport:
    """Load entry-point plugins and register returned candidates."""

    if not config:
        return PluginRegistrationReport.empty()

    if skip_if_loaded and getattr(resolver, "_oneiric_plugins_loaded", False):
        cached = getattr(resolver, "_oneiric_plugin_report", None)
        if isinstance(cached, PluginRegistrationReport):
            return cached
        return PluginRegistrationReport.empty()

    groups = _build_plugin_groups(config)
    if not groups:
        return PluginRegistrationReport.empty()

    report = _process_plugin_groups(resolver, groups)
    resolver._oneiric_plugins_loaded = True  # type: ignore[attr-defined]
    resolver._oneiric_plugin_report = report  # type: ignore[attr-defined]

    if report.registered:
        logger.info(
            "plugins-registered",
            groups=report.groups.copy(),
            registered=report.registered,
        )
    return report


def _build_plugin_groups(config: PluginsConfig) -> list[str]:
    """Build list of plugin groups to load."""
    groups: list[str] = []
    if config.auto_load:
        groups.extend(DEFAULT_ENTRY_POINT_GROUPS)
    groups.extend(config.entry_points)
    return groups


def _process_plugin_groups(
    resolver: Resolver, groups: list[str]
) -> PluginRegistrationReport:
    """Process all plugin groups and build registration report."""
    seen = set()
    report = PluginRegistrationReport()

    for group in groups:
        if group in seen:
            continue
        seen.add(group)
        report.groups.append(group)
        _process_plugin_group(resolver, group, report)

    return report


def _process_plugin_group(
    resolver: Resolver, group: str, report: PluginRegistrationReport
) -> None:
    """Process a single plugin group."""
    for result in _load_entry_point_factories(group):
        if not result.factory:
            _record_factory_error(group, result, report)
            continue

        payload, had_error = _invoke_factory(group, result, report)
        if had_error:
            # Error already recorded by _invoke_factory
            continue

        # Process payload (even if None - will be validated by _register_payload_candidates)
        _register_payload_candidates(resolver, group, result, payload, report)


def _record_factory_error(
    group: str, result: _FactoryLoadResult, report: PluginRegistrationReport
) -> None:
    """Record error for failed factory load."""
    report.errors.append(
        PluginErrorRecord(
            group=group,
            entry_point=result.entry_point,
            reason=result.error or "not-callable",
        )
    )


def _invoke_factory(
    group: str, result: _FactoryLoadResult, report: PluginRegistrationReport
) -> tuple[object | None, bool]:
    """Invoke factory and handle errors.

    Returns:
        Tuple of (payload, had_error) where had_error indicates if an exception occurred.
    """
    if result.factory is None:
        return (None, True)

    try:
        payload = result.factory()
        return (payload, False)
    except Exception as exc:  # pragma: no cover - diagnostic only
        logger.warning(
            "plugin-metadata-failed",
            group=group,
            entry_point=result.entry_point,
            error=str(exc),
        )
        report.errors.append(
            PluginErrorRecord(
                group=group, entry_point=result.entry_point, reason=str(exc)
            )
        )
        return (None, True)


def _register_payload_candidates(
    resolver: Resolver,
    group: str,
    result: _FactoryLoadResult,
    payload: object,
    report: PluginRegistrationReport,
) -> None:
    """Normalize and register candidates from payload."""
    normalized = _normalize_candidates(payload)
    if not normalized:
        report.errors.append(
            PluginErrorRecord(
                group=group,
                entry_point=result.entry_point,
                reason=f"unsupported payload {type(payload).__name__}",
            )
        )
        return

    for candidate in normalized:
        resolver.register(candidate)

    report.entries.append(
        PluginEntryRecord(
            group=group,
            entry_point=result.entry_point,
            payload_type=type(payload).__name__,
            registered_candidates=len(normalized),
        )
    )
    report.registered += len(normalized)


def _normalize_candidates(payload: object) -> list[Candidate]:
    candidates: list[Candidate] = []
    if payload is None:
        return candidates
    if isinstance(payload, Candidate):
        return [payload]
    if isinstance(payload, AdapterMetadata):
        return [payload.to_candidate()]
    if isinstance(payload, abc.Iterable) and not isinstance(payload, (str, bytes)):
        for item in payload:
            candidates.extend(_normalize_candidates(item))
        return candidates

    # Only reached if payload is not None, Candidate, AdapterMetadata, or Iterable
    logger.warning(
        "plugin-unsupported-payload",
        type=type(payload).__name__,
    )
    return candidates


def _load_entry_point_factories(group: str) -> list[_FactoryLoadResult]:
    results: list[_FactoryLoadResult] = []
    for entry_point in iter_entry_points(group):
        result = _FactoryLoadResult(group=group, entry_point=entry_point.name)
        try:
            loaded = entry_point.load()
        except Exception as exc:  # pragma: no cover - logging guard
            logger.warning(
                "plugin-load-failed",
                group=group,
                entry_point=entry_point.name,
                error=str(exc),
            )
            result.error = str(exc)
            results.append(result)
            continue
        if callable(loaded):
            result.factory = loaded
        else:
            result.error = f"not-callable:{type(loaded).__name__}"
            logger.warning(
                "plugin-not-callable",
                group=group,
                entry_point=entry_point.name,
                type=type(loaded).__name__,
            )
        results.append(result)
    return results



"""Reusable settings fragments."""

from __future__ import annotations

from pydantic import AnyHttpUrl, BaseModel, Field


class BaseURLSettings(BaseModel):
    base_url: AnyHttpUrl | None = Field(
        default=None, description="Optional base URL for outbound requests."
    )


class TimeoutSettings(BaseModel):
    timeout: float = Field(
        default=10.0, ge=0.5, description="Request timeout in seconds."
    )



"""Security, signature, and token helper actions."""

from __future__ import annotations

import base64
import hashlib
import hmac
import json
import secrets
from datetime import UTC, datetime
from typing import Any, Literal

from pydantic import BaseModel, Field

from oneiric.actions.metadata import ActionMetadata
from oneiric.actions.payloads import normalize_payload
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class SecuritySignatureSettings(BaseModel):
    """Settings controlling signature generation."""

    algorithm: Literal["sha256", "sha512", "blake2b"] = Field(
        default="sha256",
        description="Digest algorithm used when hashing the payload.",
    )
    encoding: Literal["hex", "base64"] = Field(
        default="hex",
        description="Encoding applied to the digest output.",
    )
    secret: str | None = Field(
        default=None,
        description="Default secret used when payload omits 'secret'.",
    )
    header_name: str = Field(
        default="X-Oneiric-Signature",
        description="Default header name that downstream services expect.",
    )
    include_timestamp: bool = Field(
        default=True,
        description="Include an ISO timestamp alongside the signature result.",
    )


class SecuritySignatureAction:
    """Action kit that generates deterministic HMAC signatures."""

    metadata = ActionMetadata(
        key="security.signature",
        provider="builtin-security-signature",
        factory="oneiric.actions.security:SecuritySignatureAction",
        description="Security helper generating HMAC signatures for outbound requests",
        domains=["service", "task", "workflow"],
        capabilities=["signature", "hmac", "validation"],
        stack_level=40,
        priority=400,
        source=CandidateSource.LOCAL_PKG,
        owner="Security",
        requires_secrets=True,
        side_effect_free=True,
        settings_model=SecuritySignatureSettings,
    )

    _ALGORITHMS = {
        "sha256": hashlib.sha256,
        "sha512": hashlib.sha512,
        "blake2b": hashlib.blake2b,
    }

    def __init__(self, settings: SecuritySignatureSettings | None = None) -> None:
        self._settings = settings or SecuritySignatureSettings()
        self._logger = get_logger("action.security.signature")

    async def execute(self, payload: dict | None = None) -> dict:
        payload = normalize_payload(payload)
        secret = payload.get("secret") or self._settings.secret
        if not secret:
            raise LifecycleError("security-signature-secret-required")
        algorithm = (payload.get("algorithm") or self._settings.algorithm).lower()
        if algorithm not in self._ALGORITHMS:
            raise LifecycleError("security-signature-algorithm-invalid")
        encoding = (payload.get("encoding") or self._settings.encoding).lower()
        if encoding not in {"hex", "base64"}:
            raise LifecycleError("security-signature-encoding-invalid")
        include_timestamp = payload.get("include_timestamp")
        if include_timestamp is None:
            include_timestamp = self._settings.include_timestamp
        message = self._normalize_message(payload)
        digest = hmac.new(secret.encode("utf-8"), message, self._ALGORITHMS[algorithm])
        if encoding == "hex":
            signature = digest.hexdigest()
        else:
            signature = base64.b64encode(digest.digest()).decode("ascii")
        result = {
            "status": "signed",
            "algorithm": algorithm,
            "encoding": encoding,
            "signature": signature,
            "header": payload.get("header") or self._settings.header_name,
        }
        if include_timestamp:
            result["timestamp"] = datetime.now(UTC).isoformat()
        self._logger.info(
            "security-action-signed", algorithm=algorithm, encoding=encoding
        )
        return result

    def _normalize_message(self, payload: dict) -> bytes:
        if "message" in payload and payload["message"] is not None:
            return self._to_bytes(payload["message"])
        if "body" in payload and payload["body"] is not None:
            return self._to_bytes(payload["body"])
        if "data" in payload and payload["data"] is not None:
            return self._to_bytes(payload["data"])
        raise LifecycleError("security-signature-message-required")

    def _to_bytes(self, value: Any) -> bytes:
        if isinstance(value, bytes):
            return value
        if isinstance(value, str):
            return value.encode("utf-8")
        try:
            return json.dumps(value, sort_keys=True, separators=(",", ":")).encode(
                "utf-8"
            )
        except TypeError as exc:  # pragma: no cover - serialization edge
            raise LifecycleError("security-signature-message-invalid") from exc


class SecuritySecureSettings(BaseModel):
    """Settings controlling secure token/password helpers."""

    token_length: int = Field(
        default=32, description="Default length for generated tokens."
    )
    password_iterations: int = Field(
        default=100_000,
        description="PBKDF2 iterations applied when hashing passwords.",
    )
    include_symbols: bool = Field(
        default=True,
        description="Include symbols when generating passwords.",
    )


class SecuritySecureAction:
    """Action kit that generates tokens and secures passwords."""

    metadata = ActionMetadata(
        key="security.secure",
        provider="builtin-security-secure",
        factory="oneiric.actions.security:SecuritySecureAction",
        description="Generates secure tokens/password hashes and verifies credentials",
        domains=["service", "task", "workflow"],
        capabilities=["secure", "token", "password"],
        stack_level=40,
        priority=390,
        source=CandidateSource.LOCAL_PKG,
        owner="Security",
        requires_secrets=False,
        side_effect_free=True,
        settings_model=SecuritySecureSettings,
    )

    def __init__(self, settings: SecuritySecureSettings | None = None) -> None:
        self._settings = settings or SecuritySecureSettings()
        self._logger = get_logger("action.security.secure")

    async def execute(self, payload: dict | None = None) -> dict:
        payload = normalize_payload(payload)
        mode = (payload.get("mode") or "token").lower()
        if mode == "token":
            length = int(payload.get("length") or self._settings.token_length)
            token = secrets.token_urlsafe(length)
            self._logger.info("security-action-token", length=length)
            return {"status": "token", "token": token, "length": length}
        if mode == "password-hash":
            return self._hash_password(payload)
        if mode == "password-verify":
            return self._verify_password(payload)
        if mode == "compare":
            return self._compare(payload)
        raise LifecycleError("security-secure-mode-invalid")

    def _hash_password(self, payload: dict) -> dict:
        password = payload.get("password")
        if not isinstance(password, str) or not password:
            raise LifecycleError("security-secure-password-required")
        iterations = int(
            payload.get("iterations") or self._settings.password_iterations
        )
        salt = payload.get("salt")
        if not salt:
            salt = secrets.token_hex(16)
        digest = hashlib.pbkdf2_hmac(
            "sha256",
            password.encode("utf-8"),
            salt.encode("utf-8"),
            iterations,
        ).hex()
        self._logger.info("security-action-password-hash", iterations=iterations)
        return {
            "status": "password-hash",
            "hash": digest,
            "salt": salt,
            "iterations": iterations,
        }

    def _verify_password(self, payload: dict) -> dict:
        password = payload.get("password")
        password_hash = payload.get("hash")
        salt = payload.get("salt")
        iterations = int(
            payload.get("iterations") or self._settings.password_iterations
        )
        if not all(
            isinstance(value, str) and value
            for value in (password, password_hash, salt)
        ):
            raise LifecycleError("security-secure-verification-input-invalid")
        computed = hashlib.pbkdf2_hmac(
            "sha256",
            password.encode("utf-8"),
            salt.encode("utf-8"),
            iterations,
        ).hex()
        valid = hmac.compare_digest(computed, password_hash)
        self._logger.info("security-action-password-verify", valid=valid)
        return {
            "status": "password-verify",
            "valid": valid,
        }

    def _compare(self, payload: dict) -> dict:
        first = payload.get("a")
        second = payload.get("b")
        if not isinstance(first, str) or not isinstance(second, str):
            raise LifecycleError("security-secure-compare-invalid")
        equal = hmac.compare_digest(first, second)
        self._logger.info("security-action-compare", equal=equal)
        return {
            "status": "compare",
            "equal": equal,
        }



"""Infisical secrets adapter."""

from __future__ import annotations

import asyncio
import time
from typing import Any

import httpx
from pydantic import BaseModel, Field

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class InfisicalSecretSettings(BaseModel):
    """Configuration for the Infisical adapter."""

    base_url: str = Field(
        default="https://app.infisical.com", description="Infisical API base URL."
    )
    token: str = Field(description="Machine identity token.")
    environment: str = Field(
        description="Infisical environment slug (e.g., dev, prod)."
    )
    secret_path: str = Field(default="/", description="Secret path/folder.")
    project_id: str | None = Field(
        default=None, description="Optional project identifier."
    )
    cache_ttl_seconds: int = Field(default=60, ge=5)
    http_timeout: float = Field(default=5.0, gt=0)


class InfisicalSecretAdapter:
    """Fetches secrets from Infisical with simple caching."""

    metadata = AdapterMetadata(
        category="secrets",
        provider="infisical",
        factory="oneiric.adapters.secrets.infisical:InfisicalSecretAdapter",
        capabilities=["remote", "http", "cache"],
        stack_level=40,
        priority=600,
        source=CandidateSource.LOCAL_PKG,
        owner="Security",
        requires_secrets=True,
        settings_model=InfisicalSecretSettings,
    )

    def __init__(
        self,
        settings: InfisicalSecretSettings,
        *,
        http_client: httpx.AsyncClient | None = None,
    ) -> None:
        self._settings = settings
        self._http_client = http_client
        self._owns_client = http_client is None
        self._cache: dict[str, tuple[str, float]] = {}
        self._cache_lock = asyncio.Lock()
        self._logger = get_logger("adapter.secrets.infisical").bind(
            domain="adapter",
            key="secrets",
            provider="infisical",
        )

    async def init(self) -> None:
        if not self._http_client and self._owns_client:
            self._http_client = httpx.AsyncClient(timeout=self._settings.http_timeout)
        self._logger.info("adapter-init", adapter="infisical-secrets")

    async def health(self) -> bool:
        try:
            await self.get_secret(
                "healthcheck-this-key-should-not-exist", allow_missing=True
            )
            return True
        except Exception as exc:  # pragma: no cover - network
            self._logger.warning("adapter-health-error", error=str(exc))
            return False

    async def cleanup(self) -> None:
        if self._http_client and self._owns_client:
            await self._http_client.aclose()
        self._http_client = None
        self._cache.clear()
        self._logger.info("adapter-cleanup-complete", adapter="infisical-secrets")

    async def invalidate_cache(self) -> None:
        async with self._cache_lock:
            self._cache.clear()
        self._logger.info("secrets-cache-invalidated", adapter="infisical-secrets")

    async def get_secret(self, key: str, *, allow_missing: bool = False) -> str | None:
        cached = await self._get_cached(key)
        if cached is not None:
            return cached
        client = self._ensure_client()
        url = self._settings.base_url.rstrip("/") + "/api/v3/secrets/raw"
        payload: dict[str, Any] = {
            "environment": self._settings.environment,
            "secretPath": self._settings.secret_path,
            "secretName": key,
        }
        if self._settings.project_id:
            payload["projectId"] = self._settings.project_id
        headers = {
            "Authorization": f"Bearer {self._settings.token}",
        }
        response = await client.post(url, json=payload, headers=headers)
        if response.status_code == 404 and allow_missing:
            return None
        response.raise_for_status()
        if (value := response.json().get("secretValue")) is None and not allow_missing:
            raise LifecycleError("secret-value-missing")
        if value is not None:
            await self._set_cached(key, value)
        return value

    async def _get_cached(self, key: str) -> str | None:
        async with self._cache_lock:
            cached = self._cache.get(key)
            if not cached:
                return None
            value, expires_at = cached
            if expires_at < time.monotonic():
                self._cache.pop(key, None)
                return None
            return value

    async def _set_cached(self, key: str, value: str) -> None:
        async with self._cache_lock:
            expires_at = time.monotonic() + self._settings.cache_ttl_seconds
            self._cache[key] = (value, expires_at)

    def _ensure_client(self) -> httpx.AsyncClient:
        if not self._http_client:
            raise LifecycleError("infisical-http-client-not-initialized")
        return self._http_client



"""Database adapters."""

from .duckdb import DuckDBDatabaseAdapter, DuckDBDatabaseSettings
from .mysql import MySQLDatabaseAdapter, MySQLDatabaseSettings
from .postgres import PostgresDatabaseAdapter, PostgresDatabaseSettings
from .sqlite import SQLiteDatabaseAdapter, SQLiteDatabaseSettings

__all__ = [
    "PostgresDatabaseAdapter",
    "PostgresDatabaseSettings",
    "SQLiteDatabaseAdapter",
    "SQLiteDatabaseSettings",
    "MySQLDatabaseAdapter",
    "MySQLDatabaseSettings",
    "DuckDBDatabaseAdapter",
    "DuckDBDatabaseSettings",
]



"""OpenTelemetry metrics helpers for core lifecycle + activity events."""

from __future__ import annotations

from opentelemetry import metrics

METER = metrics.get_meter("oneiric.core")

_pause_counter = METER.create_counter(
    "oneiric_activity_pause_events_total",
    unit="1",
    description="Number of pause/resume events per domain.",
)

_drain_counter = METER.create_counter(
    "oneiric_activity_drain_events_total",
    unit="1",
    description="Number of drain/clear events per domain.",
)

_swap_histogram = METER.create_histogram(
    "oneiric_lifecycle_swap_duration_ms",
    unit="ms",
    description="Lifecycle swap/activation duration in milliseconds.",
)


def record_pause_state(domain: str, paused: bool) -> None:
    """Record a pause/resume operation for the given domain."""

    _pause_counter.add(
        1, attributes={"domain": domain, "state": "paused" if paused else "resumed"}
    )


def record_drain_state(domain: str, draining: bool) -> None:
    """Record a drain/clear operation for the given domain."""

    _drain_counter.add(
        1, attributes={"domain": domain, "state": "draining" if draining else "cleared"}
    )


def record_swap_duration(
    domain: str,
    key: str,
    provider: str | None,
    duration_ms: float,
    *,
    success: bool,
) -> None:
    """Record the elapsed time for a lifecycle activation/swap."""

    attrs = {
        "domain": domain,
        "key": key,
        "provider": provider or "unknown",
        "outcome": "success" if success else "failed",
    }
    _swap_histogram.record(max(duration_ms, 0.0), attributes=attrs)



"""Serialization encode/decode action kit."""

from __future__ import annotations

import asyncio
import base64
import json
import pickle
from pathlib import Path
from typing import Any, Literal

import yaml
from pydantic import BaseModel, Field

from oneiric.actions.metadata import ActionMetadata
from oneiric.actions.payloads import normalize_payload
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class SerializationActionSettings(BaseModel):
    """Settings for the serialization action kit."""

    default_format: Literal["json", "yaml", "pickle"] = Field(
        default="json",
        description="Default serialization format when payload omits one.",
    )
    sort_keys: bool = Field(
        default=False,
        description="Sort keys for deterministic JSON/YAML output.",
    )
    ensure_ascii: bool = Field(
        default=False,
        description="Force ASCII output for JSON encoding.",
    )


class SerializationAction:
    """Action kit that encodes/decodes payloads across common formats.

    Security Warning:
        The pickle format can execute arbitrary code during deserialization.
        Only use pickle with data from trusted sources within your internal
        workflows. For external data or untrusted sources, use JSON or YAML.

    Supported Formats:
        - json: Safe for untrusted data, human-readable
        - yaml: Safe for untrusted data, more flexible than JSON
        - pickle: TRUSTED DATA ONLY - supports arbitrary Python objects
    """

    metadata = ActionMetadata(
        key="serialization.encode",
        provider="builtin-serialization",
        factory="oneiric.actions.serialization:SerializationAction",
        description="Serializes/deserializes payloads across JSON, YAML, and pickle",
        domains=["task", "service", "workflow"],
        capabilities=["encode", "decode", "serialize"],
        stack_level=30,
        priority=430,
        source=CandidateSource.LOCAL_PKG,
        owner="Platform Core",
        requires_secrets=False,
        side_effect_free=True,
        settings_model=SerializationActionSettings,
    )

    _TEXT_FORMATS = {"json", "yaml"}

    def __init__(self, settings: SerializationActionSettings | None = None) -> None:
        self._settings = settings or SerializationActionSettings()
        self._logger = get_logger("action.serialization")

    async def execute(self, payload: dict | None = None) -> dict:
        payload = normalize_payload(payload)
        mode = (payload.get("mode") or "encode").lower()
        fmt = (payload.get("format") or self._settings.default_format).lower()
        if fmt not in {"json", "yaml", "pickle"}:
            raise LifecycleError("serialization-unsupported-format")

        # Log security warning when pickle format is used
        if fmt == "pickle":
            self._logger.warning(
                "pickle-format-security-warning",
                message="Using pickle format - ensure data is from trusted sources only",
                mode=mode,
            )

        if mode == "encode":
            return await self._encode(fmt, payload)
        if mode == "decode":
            return await self._decode(fmt, payload)
        raise LifecycleError("serialization-invalid-mode")

    async def _encode(self, fmt: str, payload: dict) -> dict:  # noqa: C901
        value = payload.get("value")
        if value is None and "data" in payload:
            value = payload["data"]
        if value is None:
            raise LifecycleError("serialization-value-required")
        sort_keys = payload.get("sort_keys")
        if sort_keys is None:
            sort_keys = self._settings.sort_keys
        ensure_ascii = payload.get("ensure_ascii")
        if ensure_ascii is None:
            ensure_ascii = self._settings.ensure_ascii
        if fmt == "json":
            text = json.dumps(value, ensure_ascii=ensure_ascii, sort_keys=sort_keys)
            data_bytes = text.encode("utf-8")
        elif fmt == "yaml":
            text = yaml.safe_dump(value, sort_keys=sort_keys)
            data_bytes = text.encode("utf-8")
        else:
            # Security Note: pickle.dumps is used here for internal workflow serialization.
            # This format should ONLY be used with trusted data from internal sources.
            # For external/untrusted data, use JSON or YAML formats instead.
            text = None
            # nosemgrep: python.lang.security.deserialization.pickle.avoid-pickle
            data_bytes = pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL)
        await self._maybe_write_path(payload.get("path"), data_bytes, fmt)
        self._logger.info("serialization-action-encoded", fmt=fmt, mode="encode")
        result = {
            "status": "encoded",
            "format": fmt,
        }
        if fmt in self._TEXT_FORMATS:
            result["text"] = text if text is not None else data_bytes.decode("utf-8")
        else:
            result["encoding"] = "base64"
            result["data"] = base64.b64encode(data_bytes).decode("ascii")
        return result

    async def _decode(self, fmt: str, payload: dict) -> dict:
        raw = await self._resolve_source(fmt, payload)
        if fmt == "json":
            text = raw.decode("utf-8")
            data = json.loads(text)
        elif fmt == "yaml":
            text = raw.decode("utf-8")
            data = yaml.safe_load(text)
        else:
            # Security Note: pickle.loads can execute arbitrary code.
            # This should ONLY deserialize data from trusted internal sources.
            # Never use pickle format with data from external or untrusted sources.
            # nosemgrep: python.lang.security.deserialization.pickle.avoid-pickle
            data = pickle.loads(raw)
        self._logger.info("serialization-action-decoded", fmt=fmt, mode="decode")
        return {
            "status": "decoded",
            "format": fmt,
            "data": data,
        }

    async def _resolve_source(self, fmt: str, payload: dict) -> bytes:
        # Read from path if provided
        if path := payload.get("path"):
            return await self._read_path(path, fmt)

        # Extract value from payload
        value = self._extract_value_from_payload(payload)
        if value is None:
            raise LifecycleError("serialization-source-required")

        # Handle text formats
        if fmt in self._TEXT_FORMATS:
            return self._process_text_value(value)

        # Handle binary formats (pickle)
        return self._process_binary_value(value)

    def _extract_value_from_payload(self, payload: dict) -> Any:
        """Extract value from payload, checking multiple keys."""
        value = payload.get("data")
        if value is None and "text" in payload:
            value = payload["text"]
        if value is None and "value" in payload:
            value = payload["value"]
        return value

    def _process_text_value(self, value: Any) -> bytes:
        """Process value for text formats (JSON/YAML)."""
        if isinstance(value, bytes):
            return value
        if isinstance(value, str):
            return value.encode("utf-8")
        raise LifecycleError("serialization-text-required")

    def _process_binary_value(self, value: Any) -> bytes:
        """Process value for binary formats (pickle)."""
        if isinstance(value, bytes):
            return value
        if isinstance(value, str):
            try:
                return base64.b64decode(value.encode("ascii"))
            except Exception as exc:  # pragma: no cover - invalid base64 path
                raise LifecycleError("serialization-b64-invalid") from exc
        raise LifecycleError("serialization-binary-required")

    async def _maybe_write_path(self, path_value: Any, data: bytes, fmt: str) -> None:
        if not path_value:
            return
        path = Path(path_value)
        if fmt in self._TEXT_FORMATS:
            text = data.decode("utf-8")
            await asyncio.to_thread(path.write_text, text, encoding="utf-8")
        else:
            await asyncio.to_thread(path.write_bytes, data)

    async def _read_path(self, path_value: Any, fmt: str) -> bytes:
        path = Path(path_value)
        if fmt in self._TEXT_FORMATS:
            text = await asyncio.to_thread(path.read_text, encoding="utf-8")
            return text.encode("utf-8")
        return await asyncio.to_thread(path.read_bytes)


__all__ = ["SerializationAction", "SerializationActionSettings"]



"""Kafka queue adapter built on aiokafka."""

from __future__ import annotations

import inspect
from collections.abc import Callable, Mapping, Sequence
from typing import Any

from pydantic import BaseModel, Field, SecretStr

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class KafkaQueueSettings(BaseModel):
    """Configuration for the Kafka adapter."""

    bootstrap_servers: Sequence[str] | str = Field(
        default_factory=lambda: ["localhost:9092"],
        description="Kafka bootstrap servers.",
    )
    topic: str = Field(
        default="oneiric-events",
        description="Topic used for publish/consume operations.",
    )
    group_id: str = Field(
        default="oneiric-consumer", description="Consumer group identifier."
    )
    client_id: str = Field(default="oneiric", description="Kafka client identifier.")
    auto_offset_reset: str = Field(
        default="latest",
        description="Offset reset policy (latest or earliest).",
    )
    security_protocol: str | None = Field(
        default=None, description="Kafka security protocol (e.g., SASL_SSL)."
    )
    sasl_mechanism: str | None = Field(
        default=None, description="SASL mechanism when using SASL_* protocols."
    )
    sasl_username: str | None = Field(default=None, description="SASL username.")
    sasl_password: SecretStr | None = Field(
        default=None, description="SASL password/secret."
    )
    produce_timeout: float = Field(
        default=30.0, gt=0.0, description="Timeout (seconds) for publish operations."
    )
    consume_timeout_ms: int = Field(
        default=1000,
        ge=0,
        description="Timeout (milliseconds) for consumer getmany calls.",
    )
    consume_max_records: int = Field(
        default=50,
        ge=1,
        description="Maximum records to fetch per getmany call.",
    )


class KafkaQueueAdapter:
    """Adapter that publishes to and consumes from Kafka topics."""

    metadata = AdapterMetadata(
        category="queue",
        provider="kafka",
        factory="oneiric.adapters.queue.kafka:KafkaQueueAdapter",
        capabilities=["queue", "streaming", "fanout"],
        stack_level=20,
        priority=310,
        source=CandidateSource.LOCAL_PKG,
        owner="Messaging",
        requires_secrets=True,
        settings_model=KafkaQueueSettings,
    )

    def __init__(
        self,
        settings: KafkaQueueSettings | None = None,
        *,
        producer_factory: Callable[..., Any] | None = None,
        consumer_factory: Callable[..., Any] | None = None,
        topic_partition_factory: Callable[..., Any] | None = None,
    ) -> None:
        self._settings = settings or KafkaQueueSettings()
        self._producer_factory = producer_factory
        self._consumer_factory = consumer_factory
        self._topic_partition_factory = topic_partition_factory
        self._producer: Any | None = None
        self._consumer: Any | None = None
        self._logger = get_logger("adapter.queue.kafka").bind(
            domain="adapter",
            key="queue",
            provider="kafka",
            topic=self._settings.topic,
        )

    async def init(self) -> None:
        await self._ensure_producer()
        await self._ensure_consumer()
        self._logger.info("queue-adapter-init", provider="kafka")

    async def health(self) -> bool:
        try:
            producer = await self._ensure_producer()
            get_partitions = getattr(producer, "partitions_for", None)
            if callable(get_partitions):
                partitions = await get_partitions(self._settings.topic)
                if partitions is None:
                    raise RuntimeError("topic-not-found")
            return True
        except Exception as exc:  # pragma: no cover - network
            self._logger.warning("kafka-health-check-failed", error=str(exc))
            return False

    async def cleanup(self) -> None:
        if self._producer:
            await self._stop_component(self._producer)
            self._producer = None
        if self._consumer:
            await self._stop_component(self._consumer)
            self._consumer = None
        self._logger.info("queue-adapter-cleanup", provider="kafka")

    async def publish(
        self,
        value: bytes,
        *,
        key: bytes | None = None,
        headers: Mapping[str, bytes] | None = None,
    ) -> None:
        producer = await self._ensure_producer()
        header_list = list((headers or {}).items())
        await producer.send_and_wait(
            self._settings.topic,
            value=value,
            key=key,
            headers=header_list,
            timeout=self._settings.produce_timeout,
        )
        self._logger.debug("kafka-publish", topic=self._settings.topic, key=key)

    async def consume(self) -> list[dict[str, Any]]:
        consumer = await self._ensure_consumer()
        records = await consumer.getmany(
            timeout_ms=self._settings.consume_timeout_ms,
            max_records=self._settings.consume_max_records,
        )
        messages: list[dict[str, Any]] = []
        for tp, msgs in records.items():
            for msg in msgs:
                messages.append(
                    {
                        "topic": tp.topic,
                        "partition": tp.partition,
                        "offset": msg.offset,
                        "key": msg.key,
                        "value": msg.value,
                        "timestamp": msg.timestamp,
                        "headers": dict(msg.headers or []),
                    }
                )
        return messages

    async def commit(self, offsets: Sequence[dict[str, Any]]) -> None:
        if not offsets:
            return
        consumer = await self._ensure_consumer()
        tp_offsets = {
            self._topic_partition(item["topic"], item["partition"]): item["offset"]
            for item in offsets
        }
        await consumer.commit(offsets=tp_offsets)

    def _topic_partition(self, topic: str, partition: int) -> Any:
        if self._topic_partition_factory:
            return self._topic_partition_factory(topic, partition)
        try:
            from aiokafka.structs import (
                TopicPartition,  # type: ignore  # pragma: no cover - optional dep
            )
        except ModuleNotFoundError as exc:  # pragma: no cover - optional dep
            raise LifecycleError(
                "aiokafka-not-installed: install optional extra 'oneiric[queue-kafka]' to use KafkaQueueAdapter"
            ) from exc
        return TopicPartition(topic, partition)

    async def _ensure_producer(self) -> Any:
        if self._producer:
            return self._producer
        if self._producer_factory:
            producer = self._producer_factory(**self._producer_kwargs())
        else:
            producer = self._create_aiokafka_producer()
        await self._start_component(producer)
        self._producer = producer
        return producer

    async def _ensure_consumer(self) -> Any:
        if self._consumer:
            return self._consumer
        if self._consumer_factory:
            consumer = self._consumer_factory(**self._consumer_kwargs())
        else:
            consumer = self._create_aiokafka_consumer()
        await self._start_component(consumer)
        self._consumer = consumer
        return consumer

    def _create_aiokafka_producer(self) -> Any:
        try:
            from aiokafka import AIOKafkaProducer  # type: ignore
        except ModuleNotFoundError as exc:  # pragma: no cover - optional dep
            raise LifecycleError(
                "aiokafka-not-installed: install optional extra 'oneiric[queue-kafka]' to use KafkaQueueAdapter"
            ) from exc
        return AIOKafkaProducer(**self._producer_kwargs())

    def _create_aiokafka_consumer(self) -> Any:
        try:
            from aiokafka import AIOKafkaConsumer  # type: ignore
        except ModuleNotFoundError as exc:  # pragma: no cover - optional dep
            raise LifecycleError(
                "aiokafka-not-installed: install optional extra 'oneiric[queue-kafka]' to use KafkaQueueAdapter"
            ) from exc
        return AIOKafkaConsumer(**self._consumer_kwargs())

    def _producer_kwargs(self) -> dict[str, Any]:
        kwargs = self._security_kwargs()
        kwargs.update(
            {
                "bootstrap_servers": self._settings.bootstrap_servers,
                "client_id": self._settings.client_id,
            }
        )
        return kwargs

    def _consumer_kwargs(self) -> dict[str, Any]:
        kwargs = self._security_kwargs()
        kwargs.update(
            {
                "bootstrap_servers": self._settings.bootstrap_servers,
                "client_id": self._settings.client_id,
                "group_id": self._settings.group_id,
                "auto_offset_reset": self._settings.auto_offset_reset,
                "enable_auto_commit": False,
                "topics": [self._settings.topic],
            }
        )
        return kwargs

    def _security_kwargs(self) -> dict[str, Any]:
        kwargs: dict[str, Any] = {}
        if self._settings.security_protocol:
            kwargs["security_protocol"] = self._settings.security_protocol
        if self._settings.sasl_mechanism:
            kwargs["sasl_mechanism"] = self._settings.sasl_mechanism
        if self._settings.sasl_username:
            kwargs["sasl_plain_username"] = self._settings.sasl_username
        if self._settings.sasl_password:
            kwargs["sasl_plain_password"] = (
                self._settings.sasl_password.get_secret_value()
            )
        return kwargs

    async def _start_component(self, component: Any) -> None:
        start = getattr(component, "start", None)
        if not callable(start):
            raise LifecycleError("kafka-component-missing-start")
        result = start()
        if inspect.isawaitable(result):
            await result

    async def _stop_component(self, component: Any) -> None:
        stop = getattr(component, "stop", None)
        if not callable(stop):
            return
        result = stop()
        if inspect.isawaitable(result):
            await result



"""Runtime observability telemetry helpers."""

from __future__ import annotations

import json
from collections.abc import Mapping, Sequence
from dataclasses import asdict, dataclass
from datetime import UTC, datetime
from pathlib import Path
from typing import Any

from oneiric.core.logging import get_logger

RUNTIME_TELEMETRY_FILENAME = "runtime_telemetry.json"
telemetry_logger = get_logger("runtime.telemetry")


@dataclass
class RuntimeObservabilitySnapshot:
    """Combined event + workflow telemetry payload."""

    last_event: dict[str, Any] | None = None
    last_workflow: dict[str, Any] | None = None

    def as_dict(self) -> dict[str, Any]:
        data = asdict(self)
        if data.get("last_event") is None:
            data["last_event"] = {}
        if data.get("last_workflow") is None:
            data["last_workflow"] = {}
        return data


def load_runtime_telemetry(path: str | Path) -> RuntimeObservabilitySnapshot:
    """Load runtime telemetry snapshot from disk."""

    file = Path(path)
    if not file.exists():
        return RuntimeObservabilitySnapshot()
    try:
        data = json.loads(file.read_text())
    except Exception:
        return RuntimeObservabilitySnapshot()
    snapshot = RuntimeObservabilitySnapshot()
    if isinstance(data, Mapping):
        last_event = data.get("last_event")
        if isinstance(last_event, Mapping):
            snapshot.last_event = dict(last_event)
        last_workflow = data.get("last_workflow")
        if isinstance(last_workflow, Mapping):
            snapshot.last_workflow = dict(last_workflow)
    return snapshot


class RuntimeTelemetryRecorder:
    """Persist workflow + event telemetry for CLI inspectors and dashboards."""

    def __init__(self, target_path: str | Path) -> None:
        self._path = Path(target_path)
        self._logger = telemetry_logger

    def record_event_dispatch(self, topic: str, results: Sequence[Any]) -> None:
        snapshot = load_runtime_telemetry(self._path)
        snapshot.last_event = self._event_payload(topic, results)
        self._write(snapshot)
        payload = snapshot.last_event or {}
        self._logger.info(
            "runtime-event-telemetry",
            topic=payload.get("topic"),
            matched=payload.get("matched_handlers", 0),
            failures=payload.get("failures", 0),
        )

    def record_workflow_execution(
        self, workflow_key: str, dag_spec: Mapping[str, Any], results: Mapping[str, Any]
    ) -> None:
        snapshot = load_runtime_telemetry(self._path)
        snapshot.last_workflow = self._workflow_payload(workflow_key, dag_spec, results)
        self._write(snapshot)
        payload = snapshot.last_workflow or {}
        self._logger.info(
            "runtime-workflow-telemetry",
            workflow=payload.get("workflow"),
            node_count=payload.get("node_count", 0),
            total_duration_ms=payload.get("total_duration_ms", 0.0),
        )

    # internal -----------------------------------------------------------------

    def _event_payload(self, topic: str, results: Sequence[Any]) -> dict[str, Any]:
        handlers: list[dict[str, Any]] = []
        failures = 0
        total_duration = 0.0
        for result in results:
            duration_ms = float(getattr(result, "duration", 0.0)) * 1000.0
            total_duration += duration_ms
            success = bool(getattr(result, "success", False))
            if not success:
                failures += 1
            handlers.append(
                {
                    "handler": getattr(result, "handler", "unknown"),
                    "success": success,
                    "duration_ms": duration_ms,
                    "attempts": int(getattr(result, "attempts", 0) or 0),
                    "error": getattr(result, "error", None),
                }
            )
        return {
            "topic": topic,
            "matched_handlers": len(results),
            "failures": failures,
            "total_duration_ms": total_duration,
            "handlers": handlers,
            "recorded_at": _timestamp(),
        }

    def _workflow_payload(  # noqa: C901
        self,
        workflow_key: str,
        dag_spec: Mapping[str, Any],
        results: Mapping[str, Any],
    ) -> dict[str, Any]:
        nodes_raw = dag_spec.get("nodes") or dag_spec.get("tasks") or []
        nodes: list[dict[str, Any]] = []
        total_duration = 0.0
        for entry in nodes_raw if isinstance(nodes_raw, Sequence) else []:
            if not isinstance(entry, Mapping):
                continue
            node_id = entry.get("id") or entry.get("key")
            if not isinstance(node_id, str):
                continue
            depends = entry.get("depends_on") or []
            if not isinstance(depends, Sequence) or isinstance(depends, (str, bytes)):
                depends = []
            duration = float(results.get(f"{node_id}__duration") or 0.0)
            duration_ms = duration * 1000.0
            total_duration += duration_ms
            nodes.append(
                {
                    "node": node_id,
                    "task": entry.get("task"),
                    "depends_on": list(depends),
                    "duration_ms": duration_ms,
                    "attempts": int(results.get(f"{node_id}__attempts") or 0),
                    "retry_policy": entry.get("retry_policy") or None,
                }
            )
        entry_nodes = [node["node"] for node in nodes if not node["depends_on"]]
        return {
            "workflow": workflow_key,
            "node_count": len(nodes),
            "entry_nodes": entry_nodes,
            "nodes": nodes,
            "total_duration_ms": total_duration,
            "recorded_at": _timestamp(),
        }

    def _write(self, snapshot: RuntimeObservabilitySnapshot) -> None:
        path = self._path
        path.parent.mkdir(parents=True, exist_ok=True)
        tmp = path.with_suffix(".tmp")
        tmp.write_text(json.dumps(snapshot.as_dict()))
        tmp.replace(path)


def runtime_telemetry_path(cache_dir: str | Path) -> Path:
    """Return the default path used for runtime telemetry payloads."""

    return Path(cache_dir) / RUNTIME_TELEMETRY_FILENAME


def _timestamp() -> str:
    return datetime.now(UTC).isoformat()



"""Google Cloud Secret Manager adapter."""

from __future__ import annotations

import asyncio
import time
from pathlib import Path
from typing import Any

from pydantic import BaseModel, Field

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class GCPSecretManagerSettings(BaseModel):
    """Settings for accessing Google Cloud Secret Manager."""

    project_id: str = Field(description="Target GCP project for secrets.")
    credentials_file: Path | None = Field(
        default=None,
        description="Optional service account JSON file.",
    )
    default_version: str = Field(
        default="latest",
        description="Default version requested when callers omit a version.",
    )
    cache_ttl_seconds: int = Field(default=60, ge=0)
    healthcheck_secret: str | None = Field(
        default=None,
        description="Optional secret to probe during health checks.",
    )


class GCPSecretManagerAdapter:
    """Async adapter that fetches secrets from Google Cloud Secret Manager."""

    metadata = AdapterMetadata(
        category="secrets",
        provider="gcp-secret-manager",
        factory="oneiric.adapters.secrets.gcp:GCPSecretManagerAdapter",
        capabilities=["remote", "cache"],
        stack_level=50,
        priority=550,
        source=CandidateSource.LOCAL_PKG,
        owner="Security",
        requires_secrets=True,
        settings_model=GCPSecretManagerSettings,
    )

    def __init__(
        self,
        settings: GCPSecretManagerSettings,
        *,
        client: Any | None = None,
    ) -> None:
        self._settings = settings
        self._client = client
        self._cache: dict[str, tuple[str, float]] = {}
        self._cache_lock = asyncio.Lock()
        self._logger = get_logger("adapter.secrets.gcp").bind(
            domain="adapter",
            key="secrets",
            provider="gcp-secret-manager",
        )

    async def init(self) -> None:
        if self._client is not None:
            return
        try:
            from google.cloud import secretmanager_v1
        except ModuleNotFoundError as exc:  # pragma: no cover - defensive
            raise LifecycleError("google-secret-manager-missing") from exc
        if self._settings.credentials_file:
            self._client = secretmanager_v1.SecretManagerServiceAsyncClient.from_service_account_file(
                str(self._settings.credentials_file)
            )
        else:
            self._client = secretmanager_v1.SecretManagerServiceAsyncClient()
        self._logger.info("adapter-init", adapter="gcp-secret-manager")

    async def health(self) -> bool:
        try:
            await self.get_secret(
                self._settings.healthcheck_secret or "__oneiric_health__",
                allow_missing=True,
            )
            return True
        except Exception as exc:  # pragma: no cover - network path
            self._logger.warning("adapter-health-error", error=str(exc))
            return False

    async def cleanup(self) -> None:
        client = self._client
        self._client = None
        self._cache.clear()
        if client and hasattr(client, "close"):
            await client.close()
        self._logger.info("adapter-cleanup-complete", adapter="gcp-secret-manager")

    async def invalidate_cache(self) -> None:
        async with self._cache_lock:
            self._cache.clear()
        self._logger.info("secrets-cache-invalidated", adapter="gcp-secret-manager")

    async def get_secret(
        self,
        name: str,
        *,
        version: str | None = None,
        allow_missing: bool = False,
    ) -> str | None:
        cached = await self._get_cached(name, version)
        if cached is not None:
            return cached
        client = self._ensure_client()
        secret_name = self._format_secret_name(name, version)
        try:
            response = await client.access_secret_version(request={"name": secret_name})
        except Exception as exc:
            if allow_missing and self._is_not_found(exc):
                return None
            raise
        payload = response.payload.data.decode("utf-8")
        await self._set_cached(name, version, payload)
        return payload

    def _format_secret_name(self, secret_name: str, version: str | None) -> str:
        resolved_version = version or self._settings.default_version
        return f"projects/{self._settings.project_id}/secrets/{secret_name}/versions/{resolved_version}"

    def _ensure_client(self) -> Any:
        if not self._client:
            raise LifecycleError("gcp-secret-client-not-initialized")
        return self._client

    def _is_not_found(self, exc: Exception) -> bool:
        status = getattr(exc, "code", None)
        if status is not None and getattr(status, "name", None) == "NOT_FOUND":
            return True
        message = getattr(exc, "args", [None])[0]
        if isinstance(message, str):
            return "NOT_FOUND" in message or "404" in message
        return False

    async def _get_cached(self, name: str, version: str | None) -> str | None:
        async with self._cache_lock:
            key = self._cache_key(name, version)
            cached = self._cache.get(key)
            if not cached:
                return None
            value, expires_at = cached
            if expires_at < time.monotonic():
                self._cache.pop(key, None)
                return None
            return value

    async def _set_cached(self, name: str, version: str | None, value: str) -> None:
        async with self._cache_lock:
            key = self._cache_key(name, version)
            expires_at = time.monotonic() + self._settings.cache_ttl_seconds
            self._cache[key] = (value, expires_at)

    def _cache_key(self, name: str, version: str | None) -> str:
        return f"{name}:{version or self._settings.default_version}"



"""Remote manifest loader and artifact fetcher."""

from __future__ import annotations

import asyncio
import hashlib
import json
import tempfile
import time
from collections.abc import Sequence
from dataclasses import dataclass
from datetime import UTC, datetime
from pathlib import Path
from typing import Any
from urllib.parse import urlparse

import httpx
import yaml

from oneiric.core.config import RemoteSourceConfig, SecretsHook
from oneiric.core.logging import get_logger
from oneiric.core.resiliency import CircuitBreaker, CircuitBreakerOpen, run_with_retry
from oneiric.core.resolution import Candidate, CandidateSource, Resolver

from .metrics import (
    record_digest_checks_metric,
    record_remote_duration_metric,
    record_remote_failure_metric,
    record_remote_success_metric,
)
from .models import RemoteManifest, RemoteManifestEntry
from .security import (
    get_canonical_manifest_for_signing,
    verify_manifest_signatures,
)
from .telemetry import record_remote_failure, record_remote_success

logger = get_logger("remote")

VALID_DOMAINS = {"adapter", "action", "service", "task", "event", "workflow"}

# Default HTTP timeout for remote fetches (30 seconds)
DEFAULT_HTTP_TIMEOUT = 30.0

_REMOTE_BREAKERS: dict[str, CircuitBreaker] = {}


def _breaker_key(url: str, cache_dir: str) -> str:
    return f"{cache_dir}:{url}"


def _breaker_for(config: RemoteSourceConfig, url: str) -> CircuitBreaker:
    key = _breaker_key(url, config.cache_dir)
    breaker = _REMOTE_BREAKERS.get(key)
    if breaker is None:
        breaker = CircuitBreaker(
            name=f"remote:{url}",
            failure_threshold=config.circuit_breaker_threshold,
            recovery_time=config.circuit_breaker_reset,
            max_recovery_time=max(config.circuit_breaker_reset * 5, 0.1),
            adaptive_key=f"remote:{_remote_host(url)}",
        )
        _REMOTE_BREAKERS[key] = breaker
    return breaker


@dataclass
class RemoteSyncResult:
    manifest: RemoteManifest
    registered: int
    duration_ms: float
    per_domain: dict[str, int]
    skipped: int


def _local_path_from_url(
    url: str,
    *,
    allow_file_uris: bool,
    allowed_file_uri_roots: Sequence[str],
) -> Path | None:
    if url.startswith("file://"):
        if not allow_file_uris:
            raise ValueError("file URI access disabled for remote manifests")
        path = Path(url[7:])
        _assert_allowed_path(path, allowed_file_uri_roots)
        return path
    path = Path(url)
    if path.exists():
        if not allow_file_uris:
            raise ValueError("local manifest paths disabled for remote manifests")
        _assert_allowed_path(path, allowed_file_uri_roots)
        return path
    return None


class ArtifactManager:
    def __init__(
        self,
        cache_dir: str,
        verify_tls: bool = True,
        timeout: float = DEFAULT_HTTP_TIMEOUT,
        *,
        allow_file_uris: bool = False,
        allowed_file_uri_roots: Sequence[str] | None = None,
    ) -> None:
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.verify_tls = verify_tls
        self.timeout = timeout
        self.allow_file_uris = allow_file_uris
        self.allowed_file_uri_roots = list(allowed_file_uri_roots or [])

    async def fetch(
        self, uri: str, sha256: str | None, headers: dict[str, str]
    ) -> Path:
        """Fetch artifact with path traversal protection and timeout.

        Args:
            uri: URI to fetch (HTTP/HTTPS or local file path)
            sha256: Expected SHA256 digest (optional)
            headers: HTTP headers for request

        Returns:
            Path to cached artifact

        Raises:
            ValueError: If path traversal detected or digest mismatch
        """
        # Validate and sanitize URI
        self._validate_uri(uri)
        filename = self._get_safe_filename(uri, sha256)
        destination = self._get_destination_path(filename)

        # Return cached file if available
        if destination.exists():
            if sha256:
                _assert_digest(destination, sha256)
            return destination

        # Try local file first
        local_result = self._try_local_file(uri, destination, sha256)
        if local_result:
            return local_result

        # Fetch from remote HTTP/HTTPS
        return await self._fetch_remote_file(uri, destination, sha256, headers)

    def _validate_uri(self, uri: str) -> None:  # noqa: C901
        """Validate URI for security issues."""
        if not uri:
            raise ValueError("URI cannot be empty")
        if uri.startswith("file://") and not self.allow_file_uris:
            raise ValueError("file URI access disabled for artifacts")

        # Block path traversal attempts
        if ".." in uri and not uri.startswith(("http://", "https://", "file://")):
            raise ValueError(f"Path traversal attempt detected in URI: {uri}")

        # Block absolute paths unless they're file:// URLs
        if uri.startswith("/") and not uri.startswith("file://"):
            raise ValueError(f"Path traversal attempt detected in URI: {uri}")

        # Block relative paths with directory separators
        if not uri.startswith(("http://", "https://", "file://")):
            # Check for directory traversal characters
            if "/" in uri or "\\" in uri:
                raise ValueError(f"Path traversal attempt detected in URI: {uri}")

    def _get_safe_filename(self, uri: str, sha256: str | None) -> str:
        """Get safe filename from URI or SHA256."""
        if sha256:
            return sha256  # SHA256 is safe (hex string)

        # Sanitize filename from URI
        filename = Path(uri).name
        if "/" in filename or "\\" in filename or ".." in filename:
            raise ValueError(f"Path traversal attempt detected in URI: {uri}")
        return filename

    def _get_destination_path(self, filename: str) -> Path:
        """Get validated destination path within cache directory."""
        destination = (self.cache_dir / filename).resolve()
        cache_dir_resolved = self.cache_dir.resolve()

        # CRITICAL: Verify destination is within cache_dir
        if not destination.is_relative_to(cache_dir_resolved):
            raise ValueError(
                f"Path traversal attempt detected: {destination} "
                f"is not within cache directory {cache_dir_resolved}"
            )

        return destination

    def _try_local_file(
        self, uri: str, destination: Path, sha256: str | None
    ) -> Path | None:
        """Try to fetch from local file system."""
        # Only handle file:// URIs (absolute paths are blocked in validation)
        if not uri.startswith("file://"):
            return None
        if not self.allow_file_uris:
            raise ValueError("file URI access disabled for artifacts")

        local_path = Path(uri[7:])
        _assert_allowed_path(local_path, self.allowed_file_uri_roots)

        # Copy local file to cache
        if local_path.exists():
            data = local_path.read_bytes()
            destination.write_bytes(data)
            if sha256:
                _assert_digest(destination, sha256)
            return destination

        return None

    async def _fetch_remote_file(
        self, uri: str, destination: Path, sha256: str | None, headers: dict[str, str]
    ) -> Path:
        """Fetch file from remote HTTP/HTTPS URL."""
        # Validate URI scheme
        if not uri.startswith(("http://", "https://")):
            raise ValueError(
                f"Unsupported URI scheme (must be http://, https://, or file://): {uri}"
            )

        # Download to temporary file
        tmp_file = await self._download_to_temp_file(uri, headers)

        # Verify digest and rename
        try:
            if sha256:
                _assert_digest(tmp_file, sha256)
            tmp_file.rename(destination)
            return destination
        except Exception:
            tmp_file.unlink(missing_ok=True)
            raise

    async def _download_to_temp_file(self, uri: str, headers: dict[str, str]) -> Path:
        """Download URI to temporary file."""
        tmp_file: Path | None = None
        try:
            async with httpx.AsyncClient(
                verify=self.verify_tls,
                timeout=self.timeout,
                follow_redirects=True,
            ) as client:
                async with client.stream("GET", uri, headers=headers) as response:
                    response.raise_for_status()
                    with tempfile.NamedTemporaryFile(
                        dir=self.cache_dir,
                        prefix="dl-",
                        delete=False,
                    ) as fh:
                        tmp_file = Path(fh.name)
                        async for chunk in response.aiter_bytes():
                            fh.write(chunk)
        except Exception:
            if tmp_file is not None:
                tmp_file.unlink(missing_ok=True)
            raise

        assert tmp_file is not None  # For type checkers
        return tmp_file


async def sync_remote_manifest(
    resolver: Resolver,
    config: RemoteSourceConfig,
    *,
    secrets: SecretsHook | None = None,
    manifest_url: str | None = None,
) -> RemoteSyncResult | None:
    """Fetch a remote manifest and register its entries against the resolver."""

    url = manifest_url or config.manifest_url
    if not url:
        logger.info("remote-skip", reason="no-manifest-url")
        return None
    if not config.enabled and not manifest_url:
        logger.info("remote-skip", reason="disabled")
        return None

    breaker = _breaker_for(config, url)
    try:
        return await breaker.call(
            lambda: _run_sync(
                resolver,
                config,
                url,
                secrets,
                retry_attempts=config.max_retries,
                retry_base_delay=config.retry_base_delay,
                retry_max_delay=config.retry_max_delay,
                retry_jitter=config.retry_jitter,
            )
        )
    except CircuitBreakerOpen as exc:
        logger.warning(
            "remote-sync-circuit-open",
            url=url,
            retry_after=exc.retry_after,
        )
        return None
    except Exception as exc:
        error = str(exc)
        record_remote_failure(config.cache_dir, error)
        record_remote_failure_metric(url=url, error=error)
        raise


async def remote_sync_loop(
    resolver: Resolver,
    config: RemoteSourceConfig,
    *,
    secrets: SecretsHook | None = None,
    manifest_url: str | None = None,
    interval_override: float | None = None,
) -> None:
    """Continuously refresh remote manifest candidates based on refresh_interval."""

    url = manifest_url or config.manifest_url
    if not url:
        logger.info("remote-refresh-skip", reason="no-manifest-url")
        return
    interval = (
        interval_override if interval_override is not None else config.refresh_interval
    )
    if not interval:
        logger.info("remote-refresh-skip", reason="no-refresh-interval")
        return

    breaker = _breaker_for(config, url)
    while True:
        await asyncio.sleep(interval)
        try:
            await breaker.call(
                lambda: _run_sync(
                    resolver,
                    config,
                    url,
                    secrets,
                    retry_attempts=config.max_retries,
                    retry_base_delay=config.retry_base_delay,
                    retry_max_delay=config.retry_max_delay,
                    retry_jitter=config.retry_jitter,
                )
            )
        except CircuitBreakerOpen as exc:
            logger.warning(
                "remote-refresh-circuit-open",
                url=url,
                retry_after=exc.retry_after,
            )
            continue
        except Exception as exc:  # pragma: no cover - log and continue
            error = str(exc)
            logger.error(
                "remote-refresh-error",
                url=url,
                error=error,
            )
            record_remote_failure(config.cache_dir, error)
            record_remote_failure_metric(url=url, error=error)


async def _run_sync(
    resolver: Resolver,
    config: RemoteSourceConfig,
    url: str,
    secrets: SecretsHook | None,
    *,
    retry_attempts: int,
    retry_base_delay: float,
    retry_max_delay: float,
    retry_jitter: float,
) -> RemoteSyncResult | None:
    headers = await _auth_headers(config, secrets)
    manifest_data = await run_with_retry(
        lambda: _fetch_text(
            url,
            headers,
            verify_tls=config.verify_tls,
            allow_file_uris=config.allow_file_uris,
            allowed_file_uri_roots=config.allowed_file_uri_roots,
        ),
        attempts=retry_attempts,
        base_delay=retry_base_delay,
        max_delay=retry_max_delay,
        jitter=retry_jitter,
        adaptive_key=f"remote:manifest:{_remote_host(url)}",
        attributes=_retry_attributes(url, "manifest.fetch"),
    )
    manifest = _parse_manifest(manifest_data, signature_policy=config)
    artifact_manager = ArtifactManager(
        config.cache_dir,
        verify_tls=config.verify_tls,
        allow_file_uris=config.allow_file_uris,
        allowed_file_uri_roots=config.allowed_file_uri_roots,
    )

    registered = 0
    digest_checks = 0
    start = time.perf_counter()
    per_domain: dict[str, int] = {}
    skipped = 0
    for entry in manifest.entries:
        error = _validate_entry(entry)
        if error:
            skipped += 1
            logger.warning(
                "remote-entry-invalid",
                domain=entry.domain,
                key=entry.key,
                provider=entry.provider,
                error=error,
            )
            continue
        artifact_path = None
        if entry.uri:
            artifact_path = await run_with_retry(
                lambda: artifact_manager.fetch(entry.uri, entry.sha256, headers),
                attempts=retry_attempts,
                base_delay=retry_base_delay,
                max_delay=retry_max_delay,
                jitter=retry_jitter,
                adaptive_key=f"remote:artifact:{_remote_host(entry.uri)}",
                attributes=_retry_attributes(entry.uri, "artifact.fetch"),
            )
        if entry.sha256:
            digest_checks += 1
        candidate = _candidate_from_entry(entry, artifact_path)
        resolver.register(candidate)
        registered += 1
        per_domain[entry.domain] = per_domain.get(entry.domain, 0) + 1
    source = manifest.source or "remote"
    duration_ms = (time.perf_counter() - start) * 1000
    logger.info(
        "remote-sync-complete",
        url=url,
        registered=registered,
        source=source,
        duration_ms=duration_ms,
        digest_checks=digest_checks,
        per_domain=per_domain,
        skipped=skipped,
    )
    record_remote_success(
        config.cache_dir,
        source=source,
        registered=registered,
        duration_ms=duration_ms,
        digest_checks=digest_checks,
        per_domain=per_domain,
        skipped=skipped,
    )
    record_remote_success_metric(source=source, url=url, registered=registered)
    record_remote_duration_metric(url=url, source=source, duration_ms=duration_ms)
    record_digest_checks_metric(url=url, count=digest_checks)
    return RemoteSyncResult(
        manifest=manifest,
        registered=registered,
        duration_ms=duration_ms,
        per_domain=per_domain,
        skipped=skipped,
    )


async def _auth_headers(
    config: RemoteSourceConfig, secrets: SecretsHook | None
) -> dict[str, str]:
    token = config.auth.token
    if not token and config.auth.secret_id and secrets:
        token = await secrets.get(config.auth.secret_id)
    if not token:
        return {}
    return {config.auth.header_name: token}


async def _fetch_text(
    url: str,
    headers: dict[str, str],
    *,
    verify_tls: bool,
    allow_file_uris: bool,
    allowed_file_uri_roots: Sequence[str],
) -> str:
    local_path = _local_path_from_url(
        url,
        allow_file_uris=allow_file_uris,
        allowed_file_uri_roots=allowed_file_uri_roots,
    )
    if local_path:
        return local_path.read_text()
    if not url.startswith(("http://", "https://")):
        raise ValueError(f"Unsupported manifest URL: {url}")
    async with httpx.AsyncClient(
        verify=verify_tls,
        timeout=DEFAULT_HTTP_TIMEOUT,
        follow_redirects=True,
    ) as client:
        response = await client.get(url, headers=headers)
        response.raise_for_status()
        return response.text


def _remote_host(url: str) -> str:
    parsed = urlparse(url)
    return parsed.netloc or parsed.path.split("/")[0]


def _retry_attributes(url: str, operation: str) -> dict[str, str]:
    host = _remote_host(url) or "unknown"
    return {"domain": "remote", "operation": operation, "host": host}


def _parse_manifest(  # noqa: C901
    text: str,
    *,
    verify_signature: bool = True,
    signature_policy: RemoteSourceConfig | None = None,
) -> RemoteManifest:
    """Parse and optionally verify remote manifest.

    Args:
        text: Raw manifest text (JSON or YAML)
        verify_signature: Whether to verify signature (default: True)

    Returns:
        Parsed RemoteManifest

    Raises:
        ValueError: If manifest is invalid or signature verification fails
    """
    try:
        data = json.loads(text)
    except json.JSONDecodeError:
        data = yaml.safe_load(text)
    if not isinstance(data, dict):
        raise ValueError("Remote manifest must be a mapping at the top level.")

    policy = signature_policy or RemoteSourceConfig()

    if verify_signature:
        signatures, algorithms = _extract_signatures(data)
        if signatures:
            if any(algo != "ed25519" for algo in algorithms):
                raise ValueError("Unsupported signature algorithm in manifest.")
            canonical = get_canonical_manifest_for_signing(data)
            is_valid, error, valid_count = verify_manifest_signatures(
                canonical, signatures, threshold=policy.signature_threshold
            )
            if not is_valid:
                raise ValueError(f"Signature verification failed: {error}")
            _validate_signature_timing(data, policy)
            logger.info(
                "manifest-signature-verified",
                algorithm="ed25519",
                valid_signatures=valid_count,
                required=policy.signature_threshold,
            )
        elif policy.signature_required:
            raise ValueError("Manifest signature required but not provided.")
        else:
            logger.warning(
                "manifest-unsigned",
                recommendation="Enable signature verification for production use",
            )

    return RemoteManifest(**data)


def _extract_signatures(data: dict[str, Any]) -> tuple[list[str], list[str]]:  # noqa: C901
    signatures: list[str] = []
    algorithms: list[str] = []

    if isinstance(data.get("signatures"), list):
        for entry in data["signatures"]:
            if isinstance(entry, str):
                signatures.append(entry)
                algorithms.append("ed25519")
                continue
            if not isinstance(entry, dict):
                continue
            signature = entry.get("signature")
            if not signature:
                continue
            signatures.append(str(signature))
            algorithms.append(str(entry.get("algorithm", "ed25519")))

    if data.get("signature"):
        signatures.append(str(data["signature"]))
        algorithms.append(str(data.get("signature_algorithm", "ed25519")))

    return signatures, algorithms


def _validate_signature_timing(
    data: dict[str, Any], policy: RemoteSourceConfig
) -> None:
    signed_at = _parse_timestamp(data.get("signed_at"))
    expires_at = _parse_timestamp(data.get("expires_at"))
    now = datetime.now(UTC)

    if policy.signature_require_expiry and not expires_at:
        raise ValueError("Manifest missing expires_at while expiry is required.")

    if expires_at and now > expires_at:
        raise ValueError("Manifest signature has expired.")

    if policy.signature_max_age_seconds is not None:
        if not signed_at:
            raise ValueError("Manifest missing signed_at for age enforcement.")
        age_seconds = (now - signed_at).total_seconds()
        if age_seconds > policy.signature_max_age_seconds:
            raise ValueError("Manifest signature exceeds maximum age.")


def _parse_timestamp(value: Any) -> datetime | None:
    if not value:
        return None
    if isinstance(value, datetime):
        return value.astimezone(UTC)
    if isinstance(value, str):
        try:
            parsed = datetime.fromisoformat(value)
        except ValueError:
            raise ValueError("Invalid ISO timestamp for signature metadata.")
        if parsed.tzinfo is None:
            return parsed.replace(tzinfo=UTC)
        return parsed.astimezone(UTC)
    raise ValueError("Invalid timestamp value for signature metadata.")


def _candidate_from_entry(
    entry: RemoteManifestEntry, artifact_path: Path | None
) -> Candidate:
    """Convert manifest entry to Candidate with full metadata propagation (Stage 4 enhancement)."""
    # Build metadata dictionary
    metadata = _build_candidate_metadata(entry, artifact_path)

    # Filter out None values
    metadata = {key: value for key, value in metadata.items() if value is not None}

    return Candidate(
        domain=entry.domain,
        key=entry.key,
        provider=entry.provider,
        priority=entry.priority,
        stack_level=entry.stack_level,
        factory=entry.factory,
        metadata=metadata,
        source=CandidateSource.REMOTE_MANIFEST,
    )


def _build_candidate_metadata(
    entry: RemoteManifestEntry, artifact_path: Path | None
) -> dict[str, Any]:
    """Build complete metadata dictionary for candidate."""
    metadata = entry.metadata.copy()

    # Add core metadata
    _add_core_metadata(metadata, entry, artifact_path)

    # Add domain-specific metadata
    _add_adapter_metadata(metadata, entry)
    _add_action_metadata(metadata, entry)
    _add_dependency_metadata(metadata, entry)
    _add_platform_metadata(metadata, entry)
    _add_documentation_metadata(metadata, entry)
    _add_event_metadata(metadata, entry)
    _add_dag_metadata(metadata, entry)

    return metadata


def _add_core_metadata(
    metadata: dict[str, Any], entry: RemoteManifestEntry, artifact_path: Path | None
) -> None:
    """Add core metadata fields."""
    metadata.update(
        {
            "remote_uri": entry.uri,
            "artifact_path": str(artifact_path) if artifact_path else None,
            "version": entry.version,
            "source": "remote",
        }
    )


def _add_adapter_metadata(metadata: dict[str, Any], entry: RemoteManifestEntry) -> None:
    """Add adapter-specific metadata."""
    if entry.capabilities:
        metadata["capabilities"] = entry.capability_names
        metadata["capability_descriptors"] = entry.capability_payloads()
    if entry.owner:
        metadata["owner"] = entry.owner
    metadata["requires_secrets"] = entry.requires_secrets
    if entry.settings_model:
        metadata["settings_model"] = entry.settings_model


def _add_action_metadata(metadata: dict[str, Any], entry: RemoteManifestEntry) -> None:
    """Add action-specific metadata."""
    metadata["side_effect_free"] = entry.side_effect_free
    if entry.timeout_seconds is not None:
        metadata["timeout_seconds"] = entry.timeout_seconds
    if entry.retry_policy:
        metadata["retry_policy"] = entry.retry_policy


def _add_dependency_metadata(
    metadata: dict[str, Any], entry: RemoteManifestEntry
) -> None:
    """Add dependency constraint metadata."""
    if entry.requires:
        metadata["requires"] = entry.requires
    if entry.conflicts_with:
        metadata["conflicts_with"] = entry.conflicts_with


def _add_platform_metadata(
    metadata: dict[str, Any], entry: RemoteManifestEntry
) -> None:
    """Add platform constraint metadata."""
    if entry.python_version:
        metadata["python_version"] = entry.python_version
    if entry.os_platform:
        metadata["os_platform"] = entry.os_platform


def _add_documentation_metadata(
    metadata: dict[str, Any], entry: RemoteManifestEntry
) -> None:
    """Add documentation metadata."""
    if entry.license:
        metadata["license"] = entry.license
    if entry.documentation_url:
        metadata["documentation_url"] = entry.documentation_url


def _add_event_metadata(metadata: dict[str, Any], entry: RemoteManifestEntry) -> None:
    """Add event-specific metadata."""
    if entry.event_topics:
        metadata["topics"] = entry.event_topics
    if entry.event_max_concurrency is not None:
        metadata["max_concurrency"] = entry.event_max_concurrency
    if entry.event_filters:
        metadata["filters"] = entry.event_filters
    if entry.event_priority is not None:
        metadata["event_priority"] = entry.event_priority
    if entry.event_fanout_policy:
        metadata["fanout_policy"] = entry.event_fanout_policy


def _add_dag_metadata(metadata: dict[str, Any], entry: RemoteManifestEntry) -> None:
    """Add DAG metadata."""
    if entry.dag:
        metadata["dag"] = entry.dag


def _assert_digest(path: Path, expected: str) -> None:
    digest = hashlib.sha256(path.read_bytes()).hexdigest()
    if digest != expected.lower():
        raise ValueError(
            f"Digest mismatch for {path}: expected {expected}, got {digest}"
        )


def _assert_allowed_path(path: Path, allowed_roots: Sequence[str]) -> None:
    if not allowed_roots:
        return
    resolved = path.expanduser().resolve()
    for root in allowed_roots:
        root_path = Path(root).expanduser().resolve()
        if resolved.is_relative_to(root_path):
            return
    raise ValueError(f"Local path access denied: {path}")


def _validate_entry(entry: RemoteManifestEntry) -> str | None:
    """Comprehensive validation of remote manifest entry.

    Validates domain, key format, provider, factory format, and bounds.
    """
    from oneiric.core.security import (
        validate_factory_string,
        validate_key_format,
        validate_priority_bounds,
        validate_stack_level_bounds,
    )

    # Validate basic fields
    if error := _validate_domain(entry):
        return error
    if error := _validate_key_field(entry, validate_key_format):
        return error
    if error := _validate_provider_field(entry, validate_key_format):
        return error
    if error := _validate_factory_field(entry, validate_factory_string):
        return error

    # Validate bounds
    if error := _validate_bounds(
        entry, validate_priority_bounds, validate_stack_level_bounds
    ):
        return error

    # Validate URI
    if error := _validate_uri_field(entry):
        return error

    return None


def _validate_domain(entry: RemoteManifestEntry) -> str | None:
    """Validate domain field."""
    if entry.domain not in VALID_DOMAINS:
        return f"unsupported domain '{entry.domain}'"
    return None


def _validate_key_field(entry: RemoteManifestEntry, validate_key_format) -> str | None:
    """Validate key field."""
    if not entry.key:
        return "missing key"
    is_valid, error = validate_key_format(entry.key)
    if not is_valid:
        return f"invalid key: {error}"
    return None


def _validate_provider_field(
    entry: RemoteManifestEntry, validate_key_format
) -> str | None:
    """Validate provider field."""
    if not entry.provider:
        return "missing provider"
    is_valid, error = validate_key_format(entry.provider)
    if not is_valid:
        return f"invalid provider: {error}"
    return None


def _validate_factory_field(
    entry: RemoteManifestEntry, validate_factory_string
) -> str | None:
    """Validate factory field."""
    if not entry.factory:
        return "missing factory"
    is_valid, error = validate_factory_string(entry.factory)
    if not is_valid:
        return f"invalid factory: {error}"
    return None


def _validate_bounds(
    entry: RemoteManifestEntry, validate_priority_bounds, validate_stack_level_bounds
) -> str | None:
    """Validate priority and stack level bounds."""
    # Priority bounds
    if entry.priority is not None:
        is_valid, error = validate_priority_bounds(entry.priority)
        if not is_valid:
            return error

    # Stack level bounds
    if entry.stack_level is not None:
        is_valid, error = validate_stack_level_bounds(entry.stack_level)
        if not is_valid:
            return error

    return None


def _validate_uri_field(entry: RemoteManifestEntry) -> str | None:
    """Validate URI field for path traversal."""
    if entry.uri and entry.uri.startswith(".."):
        return f"URI contains path traversal: {entry.uri}"
    return None



"""HTTPS upload adapter."""

from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path

import httpx
from pydantic import AnyHttpUrl, BaseModel, Field, SecretStr

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class HTTPSUploadSettings(BaseModel):
    """Configuration for HTTPS uploads."""

    base_url: AnyHttpUrl | None = Field(
        default=None,
        description="Optional base URL prepended to relative upload paths.",
    )
    method: str = Field(
        default="PUT",
        description="HTTP method used for uploads (PUT or POST).",
    )
    timeout: float = Field(default=30.0, ge=1.0)
    verify_tls: bool = True
    default_headers: dict[str, str] = Field(default_factory=dict)
    auth_token: SecretStr | None = Field(
        default=None,
        description="Optional token injected into Authorization headers.",
    )
    auth_scheme: str = Field(
        default="Bearer",
        description="Prefix used for the Authorization header (e.g., Bearer).",
    )
    auth_header: str = Field(
        default="Authorization",
        description="Header used for auth tokens.",
    )


class HTTPSUploadAdapter:
    """Upload artifacts over HTTPS using httpx."""

    metadata = AdapterMetadata(
        category="file_transfer",
        provider="https-upload",
        factory="oneiric.adapters.file_transfer.http_upload:HTTPSUploadAdapter",
        capabilities=["upload"],
        stack_level=15,
        priority=360,
        source=CandidateSource.LOCAL_PKG,
        owner="Platform Core",
        requires_secrets=False,
        settings_model=HTTPSUploadSettings,
    )

    def __init__(
        self,
        settings: HTTPSUploadSettings,
        *,
        client: httpx.AsyncClient | None = None,
    ) -> None:
        self._settings = settings
        self._client = client
        self._owns_client = client is None
        self._logger = get_logger("adapter.file_transfer.https_upload").bind(
            domain="adapter",
            key="file_transfer",
            provider="https-upload",
        )

    async def init(self) -> None:
        if self._client is None:
            base_url = (
                str(self._settings.base_url)
                if self._settings.base_url is not None
                else httpx.URL("")
            )
            self._client = httpx.AsyncClient(
                timeout=self._settings.timeout,
                verify=self._settings.verify_tls,
                base_url=base_url,
            )
            self._owns_client = True
        self._logger.info("https-upload-init")

    async def cleanup(self) -> None:
        if self._client and self._owns_client:
            await self._client.aclose()
        self._client = None
        self._logger.info("https-upload-cleanup")

    async def health(self) -> bool:
        if self._settings.base_url is None:
            return True
        try:
            response = await self._ensure_client().get("", timeout=5.0)
            return response.status_code < 500
        except httpx.HTTPError:  # pragma: no cover - network path
            return False

    async def upload(
        self,
        path_or_url: str,
        data: bytes,
        *,
        content_type: str | None = None,
        extra_headers: Mapping[str, str] | None = None,
    ) -> str:
        client = self._ensure_client()
        headers = self._settings.default_headers.copy()
        if self._settings.auth_token:
            token = self._settings.auth_token.get_secret_value()
            scheme = self._settings.auth_scheme
            prefix = f"{scheme} " if scheme else ""
            headers.setdefault(self._settings.auth_header, f"{prefix}{token}")
        if content_type:
            headers.setdefault("Content-Type", content_type)
        if extra_headers:
            headers.update(extra_headers)

        request_target = path_or_url
        if self._settings.base_url and not path_or_url.startswith("http"):
            request_target = path_or_url.lstrip("/")

        try:
            response = await client.request(
                self._settings.method.upper(),
                request_target,
                content=data,
                headers=headers,
            )
            response.raise_for_status()
        except httpx.HTTPError as exc:
            self._logger.error(
                "https-upload-failed", url=request_target, error=str(exc)
            )
            raise LifecycleError("https-upload-failed") from exc

        location = response.headers.get("Location") or str(response.request.url)
        return location

    async def upload_file(
        self,
        path_or_url: str,
        file_path: str | Path,
        *,
        content_type: str | None = None,
        extra_headers: Mapping[str, str] | None = None,
    ) -> str:
        data = Path(file_path).read_bytes()
        return await self.upload(
            path_or_url,
            data,
            content_type=content_type,
            extra_headers=extra_headers,
        )

    def _ensure_client(self) -> httpx.AsyncClient:
        if not self._client:
            raise LifecycleError("https-upload-client-not-initialized")
        return self._client



"""OpenAI LLM adapter for GPT models."""

from __future__ import annotations

import asyncio
import json
from collections.abc import AsyncGenerator
from typing import Any

from pydantic import Field, SecretStr

from oneiric.adapters.llm.common import (
    LLMBase,
    LLMBaseSettings,
    LLMCapability,
    LLMMessage,
    LLMModelInfo,
    LLMProvider,
    LLMResponse,
    LLMStreamChunk,
)
from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class OpenAILLMSettings(LLMBaseSettings):
    """Settings for OpenAI LLM adapter."""

    # OpenAI-specific settings
    openai_api_key: SecretStr | None = Field(default=None)
    openai_organization: str | None = Field(default=None)
    openai_base_url: str = Field(default="https://api.openai.com/v1")

    # Model defaults
    model: str = Field(default="gpt-3.5-turbo")

    # Advanced OpenAI settings
    user: str | None = Field(default=None)
    logprobs: bool | None = Field(default=None)
    top_logprobs: int | None = Field(default=None, ge=0, le=5)


class OpenAILLMAdapter(LLMBase):
    """OpenAI LLM adapter implementation."""

    metadata = AdapterMetadata(
        category="llm",
        provider="openai",
        factory="oneiric.adapters.llm.openai:OpenAILLMAdapter",
        description="OpenAI GPT adapter with chat, streaming, tool-calling, and JSON mode support",
        capabilities=[
            "chat_completion",
            "streaming",
            "function_calling",
            "tool_use",
            "json_mode",
            "vision",
        ],
        stack_level=20,
        priority=500,
        source=CandidateSource.LOCAL_PKG,
        owner="AI Platform",
        requires_secrets=True,
        settings_model=OpenAILLMSettings,
    )

    def __init__(
        self,
        openai_api_key: str | SecretStr | None = None,
        openai_organization: str | None = None,
        openai_base_url: str | None = None,
        model: str = "gpt-3.5-turbo",
        max_tokens: int = 1000,
        temperature: float = 0.7,
        timeout: float = 60.0,
        max_retries: int = 3,
        **kwargs: Any,
    ) -> None:
        # Convert string api_key to SecretStr
        if isinstance(openai_api_key, str):
            openai_api_key = SecretStr(openai_api_key)

        settings = OpenAILLMSettings(
            openai_api_key=openai_api_key,
            openai_organization=openai_organization,
            openai_base_url=openai_base_url or "https://api.openai.com/v1",
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
            timeout=timeout,
            max_retries=max_retries,
            **kwargs,
        )
        super().__init__(settings)
        self._logger = get_logger("adapter.llm.openai")

    @property
    def settings(self) -> OpenAILLMSettings:
        """Get adapter settings with correct type."""
        return self._settings  # type: ignore[return-value]

    async def init(self) -> None:
        """Initialize the OpenAI LLM adapter."""
        self._logger.info("Initializing OpenAI LLM adapter")
        await self._ensure_client()
        self._logger.info("OpenAI LLM adapter initialized successfully")

    async def health(self) -> bool:
        """Check if OpenAI service is healthy."""
        try:
            client = await self._ensure_client()
            # Simple models list call to verify connectivity
            await client.models.list()
            return True
        except Exception as e:
            self._logger.error(f"Health check failed: {e}")
            return False

    async def cleanup(self) -> None:
        """Cleanup OpenAI adapter resources."""
        self._logger.info("Cleaning up OpenAI LLM adapter")
        if self._client:
            await self._client.close()
            self._client = None
        self._logger.info("OpenAI LLM adapter cleanup complete")

    async def _ensure_client(self) -> Any:
        """Ensure OpenAI client is initialized."""
        if self._client is not None:
            return self._client

        try:
            import openai
        except ImportError as e:
            msg = "openai package required for OpenAI LLM adapter"
            raise ImportError(msg) from e

        # Get API key from settings
        api_key = self.settings.openai_api_key or self.settings.api_key
        if not api_key:
            msg = "OpenAI API key is required"
            raise ValueError(msg)

        self._client = openai.AsyncOpenAI(
            api_key=api_key.get_secret_value(),
            organization=self.settings.openai_organization
            or self.settings.organization,
            base_url=self.settings.openai_base_url or self.settings.base_url,
            timeout=self.settings.timeout,
            max_retries=self.settings.max_retries,
        )

        self._logger.info(
            "OpenAI client initialized",
            extra={
                "model": self.settings.model,
                "base_url": self.settings.openai_base_url,
            },
        )

        return self._client

    async def _chat(
        self,
        messages: list[LLMMessage | dict[str, Any]],
        model: str,
        temperature: float,
        max_tokens: int,
        stream: bool,
        **kwargs: Any,
    ) -> LLMResponse:
        """Generate chat completion using OpenAI."""
        client = await self._ensure_client()
        normalized_messages = self._normalize_messages(messages)

        start_time = asyncio.get_event_loop().time()

        # Build request parameters
        params = self._build_openai_request_params(
            normalized_messages, model, temperature, max_tokens, **kwargs
        )

        try:
            response = await client.chat.completions.create(**params)
            latency_ms = int((asyncio.get_event_loop().time() - start_time) * 1000)

            # Extract response data
            choice = response.choices[0]
            message = choice.message
            function_calls = self._extract_function_calls(message)

            # Get token usage and cost
            prompt_tokens, completion_tokens, total_tokens = (
                self._extract_openai_token_usage(response)
            )
            model_info = await self._get_model_info(model)
            cost = self._calculate_cost(prompt_tokens, completion_tokens, model_info)

            return LLMResponse(
                content=message.content or "",
                model=response.model,
                provider=LLMProvider.OPENAI.value,
                finish_reason=choice.finish_reason,
                tokens_used=total_tokens,
                prompt_tokens=prompt_tokens,
                completion_tokens=completion_tokens,
                latency_ms=latency_ms,
                function_calls=function_calls,
                cost=cost,
                metadata={
                    "system_fingerprint": getattr(response, "system_fingerprint", None),
                    "created": response.created,
                },
            )

        except Exception as e:
            self._logger.error(f"Chat completion failed: {e}", exc_info=True)
            raise

    def _build_openai_request_params(
        self,
        normalized_messages: list[dict[str, Any]],
        model: str,
        temperature: float,
        max_tokens: int,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Build request parameters for OpenAI API."""
        params: dict[str, Any] = {
            "model": model,
            "messages": normalized_messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "stream": False,
        }

        # Add optional parameters from settings
        self._add_optional_settings_params(params)

        # Add function/tool calling parameters
        self._add_function_tool_params(params, kwargs)

        return params

    def _add_optional_settings_params(self, params: dict[str, Any]) -> None:
        """Add optional parameters from settings to request."""
        self._add_numeric_params(params)
        self._add_conditional_params(params)

    def _add_numeric_params(self, params: dict[str, Any]) -> None:
        """Add numeric parameters that differ from defaults."""
        optional_settings = {
            "top_p": (self.settings.top_p, 1.0),
            "frequency_penalty": (self.settings.frequency_penalty, 0.0),
            "presence_penalty": (self.settings.presence_penalty, 0.0),
            "n": (self.settings.n, 1),
        }
        for param_name, (value, default) in optional_settings.items():
            if value != default:
                params[param_name] = value

    def _add_conditional_params(self, params: dict[str, Any]) -> None:
        """Add parameters that are set (non-empty/non-None)."""
        conditional_params = [
            ("stop", self.settings.stop),
            ("logit_bias", self.settings.logit_bias),
            ("user", self.settings.user),
            ("seed", self.settings.seed),
            ("response_format", self.settings.response_format),
            ("logprobs", self.settings.logprobs),
            ("top_logprobs", self.settings.top_logprobs),
        ]
        for param_name, value in conditional_params:
            if value:
                params[param_name] = value

    def _add_function_tool_params(
        self, params: dict[str, Any], kwargs: dict[str, Any]
    ) -> None:
        """Add function and tool calling parameters to request."""
        if kwargs.get("functions"):
            params["functions"] = kwargs["functions"]
        if kwargs.get("function_call"):
            params["function_call"] = kwargs["function_call"]
        if kwargs.get("tools"):
            params["tools"] = kwargs["tools"]
        if kwargs.get("tool_choice"):
            params["tool_choice"] = kwargs["tool_choice"]

    def _extract_function_calls(self, message: Any) -> list[dict[str, Any]] | None:
        """Extract function calls from OpenAI message."""
        if hasattr(message, "function_call") and message.function_call:
            return [
                {
                    "name": message.function_call.name,
                    "arguments": json.loads(message.function_call.arguments),
                }
            ]
        elif hasattr(message, "tool_calls") and message.tool_calls:
            return [
                {
                    "id": call.id,
                    "type": call.type,
                    "name": call.function.name,
                    "arguments": json.loads(call.function.arguments),
                }
                for call in message.tool_calls
            ]
        return None

    def _extract_openai_token_usage(self, response: Any) -> tuple[int, int, int]:
        """Extract token usage from OpenAI response."""
        prompt_tokens = response.usage.prompt_tokens if response.usage else 0
        completion_tokens = response.usage.completion_tokens if response.usage else 0
        total_tokens = response.usage.total_tokens if response.usage else 0
        return prompt_tokens, completion_tokens, total_tokens

    async def _chat_stream(
        self,
        messages: list[LLMMessage | dict[str, Any]],
        model: str,
        temperature: float,
        max_tokens: int,
        **kwargs: Any,
    ) -> AsyncGenerator[LLMStreamChunk]:
        """Generate streaming chat completion using OpenAI."""
        client = await self._ensure_client()
        normalized_messages = self._normalize_messages(messages)

        # Build request parameters
        params = self._build_stream_request_params(
            normalized_messages, model, temperature, max_tokens, **kwargs
        )

        try:
            stream = await client.chat.completions.create(**params)
            async for chunk in self._process_stream_chunks(stream, model):
                yield chunk
        except Exception as e:
            self._logger.error(f"Streaming chat completion failed: {e}", exc_info=True)
            raise

    def _build_stream_request_params(
        self,
        normalized_messages: list[dict[str, Any]],
        model: str,
        temperature: float,
        max_tokens: int,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Build request parameters for streaming completion."""
        params: dict[str, Any] = {
            "model": model,
            "messages": normalized_messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "stream": True,
        }

        # Add optional parameters
        self._add_stream_settings_params(params)
        self._add_stream_function_params(params, kwargs)

        return params

    def _add_stream_settings_params(self, params: dict[str, Any]) -> None:
        """Add optional settings parameters for streaming."""
        if self.settings.top_p != 1.0:
            params["top_p"] = self.settings.top_p
        if self.settings.frequency_penalty != 0.0:
            params["frequency_penalty"] = self.settings.frequency_penalty
        if self.settings.presence_penalty != 0.0:
            params["presence_penalty"] = self.settings.presence_penalty
        if self.settings.stop:
            params["stop"] = self.settings.stop
        if self.settings.user:
            params["user"] = self.settings.user
        if self.settings.seed:
            params["seed"] = self.settings.seed

    def _add_stream_function_params(
        self, params: dict[str, Any], kwargs: dict[str, Any]
    ) -> None:
        """Add function/tool calling parameters for streaming."""
        if kwargs.get("functions"):
            params["functions"] = kwargs["functions"]
        if kwargs.get("tools"):
            params["tools"] = kwargs["tools"]

    async def _process_stream_chunks(
        self, stream: Any, model: str
    ) -> AsyncGenerator[LLMStreamChunk]:
        """Process streaming response chunks."""
        async for chunk in stream:
            if not chunk.choices:
                continue

            choice = chunk.choices[0]
            delta = choice.delta

            if not delta.content:
                continue

            yield LLMStreamChunk(
                content=delta.content,
                finish_reason=choice.finish_reason,
                model=chunk.model,
                delta=True,
            )

    async def _function_call(
        self,
        messages: list[LLMMessage | dict[str, Any]],
        functions: list[dict[str, Any]],
        model: str,
        **kwargs: Any,
    ) -> LLMResponse:
        """Generate chat completion with function calling."""
        return await self._chat(
            messages=messages,
            model=model,
            temperature=self.settings.temperature,
            max_tokens=self.settings.max_tokens,
            stream=False,
            functions=functions,
            function_call=kwargs.get("function_call", "auto"),
        )

    async def _tool_use(
        self,
        messages: list[LLMMessage | dict[str, Any]],
        tools: list[dict[str, Any]],
        model: str,
        **kwargs: Any,
    ) -> LLMResponse:
        """Generate chat completion with tool use."""
        return await self._chat(
            messages=messages,
            model=model,
            temperature=self.settings.temperature,
            max_tokens=self.settings.max_tokens,
            stream=False,
            tools=tools,
            tool_choice=kwargs.get("tool_choice", "auto"),
        )

    async def _get_model_info(self, model: str) -> LLMModelInfo:
        """Get information about an OpenAI model."""
        if model in self._model_cache:
            return self._model_cache[model]

        # Static model information (OpenAI doesn't provide this via API)
        model_data = _OPENAI_MODEL_DATA.get(model)
        if not model_data:
            # Default fallback for unknown models
            model_data = {
                "context_length": 4096,
                "max_output_tokens": 4096,
                "capabilities": [LLMCapability.CHAT_COMPLETION],
                "supports_functions": False,
                "supports_tools": False,
                "supports_vision": False,
                "supports_json_mode": False,
                "cost_per_1k_input_tokens": None,
                "cost_per_1k_output_tokens": None,
            }

        model_info = LLMModelInfo(
            name=model,
            provider=LLMProvider.OPENAI,
            capabilities=model_data["capabilities"],
            context_length=model_data["context_length"],
            max_output_tokens=model_data["max_output_tokens"],
            supports_streaming=True,
            supports_functions=model_data["supports_functions"],
            supports_tools=model_data["supports_tools"],
            supports_vision=model_data["supports_vision"],
            supports_json_mode=model_data["supports_json_mode"],
            cost_per_1k_input_tokens=model_data["cost_per_1k_input_tokens"],
            cost_per_1k_output_tokens=model_data["cost_per_1k_output_tokens"],
        )

        self._model_cache[model] = model_info
        return model_info

    async def _list_models(self) -> list[LLMModelInfo]:
        """List available OpenAI models."""
        models = []
        for model_name in _OPENAI_MODEL_DATA.keys():
            model_info = await self._get_model_info(model_name)
            models.append(model_info)
        return models

    async def _count_tokens(self, text: str, model: str) -> int:
        """Count tokens using tiktoken."""
        try:
            import tiktoken

            encoding = tiktoken.encoding_for_model(model)
            return len(encoding.encode(text))
        except ImportError:
            # Fallback to rough estimation
            return await super()._count_tokens(text, model)
        except Exception:
            # If model not found in tiktoken, use rough estimation
            return await super()._count_tokens(text, model)


# OpenAI model data (as of January 2025)
_OPENAI_MODEL_DATA: dict[str, dict[str, Any]] = {
    # GPT-4 Turbo
    "gpt-4-turbo": {
        "context_length": 128000,
        "max_output_tokens": 4096,
        "capabilities": [
            LLMCapability.CHAT_COMPLETION,
            LLMCapability.FUNCTION_CALLING,
            LLMCapability.TOOL_USE,
            LLMCapability.VISION,
            LLMCapability.JSON_MODE,
        ],
        "supports_functions": True,
        "supports_tools": True,
        "supports_vision": True,
        "supports_json_mode": True,
        "cost_per_1k_input_tokens": 0.01,
        "cost_per_1k_output_tokens": 0.03,
    },
    "gpt-4-turbo-preview": {
        "context_length": 128000,
        "max_output_tokens": 4096,
        "capabilities": [
            LLMCapability.CHAT_COMPLETION,
            LLMCapability.FUNCTION_CALLING,
            LLMCapability.TOOL_USE,
            LLMCapability.JSON_MODE,
        ],
        "supports_functions": True,
        "supports_tools": True,
        "supports_vision": False,
        "supports_json_mode": True,
        "cost_per_1k_input_tokens": 0.01,
        "cost_per_1k_output_tokens": 0.03,
    },
    # GPT-4
    "gpt-4": {
        "context_length": 8192,
        "max_output_tokens": 8192,
        "capabilities": [
            LLMCapability.CHAT_COMPLETION,
            LLMCapability.FUNCTION_CALLING,
            LLMCapability.TOOL_USE,
        ],
        "supports_functions": True,
        "supports_tools": True,
        "supports_vision": False,
        "supports_json_mode": False,
        "cost_per_1k_input_tokens": 0.03,
        "cost_per_1k_output_tokens": 0.06,
    },
    "gpt-4-32k": {
        "context_length": 32768,
        "max_output_tokens": 32768,
        "capabilities": [
            LLMCapability.CHAT_COMPLETION,
            LLMCapability.FUNCTION_CALLING,
            LLMCapability.TOOL_USE,
        ],
        "supports_functions": True,
        "supports_tools": True,
        "supports_vision": False,
        "supports_json_mode": False,
        "cost_per_1k_input_tokens": 0.06,
        "cost_per_1k_output_tokens": 0.12,
    },
    # GPT-3.5 Turbo
    "gpt-3.5-turbo": {
        "context_length": 16385,
        "max_output_tokens": 4096,
        "capabilities": [
            LLMCapability.CHAT_COMPLETION,
            LLMCapability.FUNCTION_CALLING,
            LLMCapability.TOOL_USE,
            LLMCapability.JSON_MODE,
        ],
        "supports_functions": True,
        "supports_tools": True,
        "supports_vision": False,
        "supports_json_mode": True,
        "cost_per_1k_input_tokens": 0.0005,
        "cost_per_1k_output_tokens": 0.0015,
    },
    "gpt-3.5-turbo-16k": {
        "context_length": 16385,
        "max_output_tokens": 4096,
        "capabilities": [
            LLMCapability.CHAT_COMPLETION,
            LLMCapability.FUNCTION_CALLING,
            LLMCapability.TOOL_USE,
        ],
        "supports_functions": True,
        "supports_tools": True,
        "supports_vision": False,
        "supports_json_mode": False,
        "cost_per_1k_input_tokens": 0.003,
        "cost_per_1k_output_tokens": 0.004,
    },
    # GPT-4 Vision
    "gpt-4-vision-preview": {
        "context_length": 128000,
        "max_output_tokens": 4096,
        "capabilities": [
            LLMCapability.CHAT_COMPLETION,
            LLMCapability.VISION,
        ],
        "supports_functions": False,
        "supports_tools": False,
        "supports_vision": True,
        "supports_json_mode": False,
        "cost_per_1k_input_tokens": 0.01,
        "cost_per_1k_output_tokens": 0.03,
    },
}



"""Remote manifest loading utilities."""

from .loader import RemoteSyncResult, remote_sync_loop, sync_remote_manifest
from .models import RemoteManifest, RemoteManifestEntry
from .telemetry import RemoteSyncTelemetry, load_remote_telemetry

__all__ = [
    "RemoteManifest",
    "RemoteManifestEntry",
    "RemoteSyncResult",
    "sync_remote_manifest",
    "remote_sync_loop",
    "RemoteSyncTelemetry",
    "load_remote_telemetry",
]



"""HTTP adapter backed by httpx.Client (supports async shim when needed)."""

from __future__ import annotations

import asyncio
import inspect
from typing import Any

import httpx
from pydantic import Field

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.http_helpers import observed_http_request
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.observability import inject_trace_context
from oneiric.core.resolution import CandidateSource
from oneiric.core.settings_mixins import BaseURLSettings


class HTTPClientSettings(BaseURLSettings):
    timeout: float = Field(
        default=10.0,
        ge=0.1,
        description="Request timeout in seconds.",
    )
    verify: bool = Field(
        default=True,
        description="Whether to verify TLS certificates.",
    )
    headers: dict[str, str] = Field(
        default_factory=dict,
        description="Default headers merged into each request.",
    )
    healthcheck_path: str = Field(
        default="/",
        description="Relative path to hit during health checks when base_url is configured.",
    )


class _AsyncClientShim:
    """Async-friendly wrapper for httpx.Client when AsyncClient is unavailable."""

    def __init__(
        self,
        *,
        base_url: str | None,
        timeout: float,
        verify: bool,
        headers: dict[str, str] | None,
        transport: Any,
    ) -> None:
        client_kwargs: dict[str, Any] = {}
        if base_url:
            client_kwargs["url"] = base_url
        if headers:
            client_kwargs["headers"] = headers
        if transport:
            client_kwargs["transport"] = transport

        self._client = httpx.Client(**client_kwargs)
        self._timeout = timeout
        self._verify = verify

    async def request(self, method: str, url: str, **kwargs: Any) -> httpx.Response:
        return await asyncio.to_thread(self._client.request, method, url, **kwargs)

    async def get(self, url: str, **kwargs: Any) -> httpx.Response:
        return await asyncio.to_thread(self._client.get, url, **kwargs)

    async def post(self, url: str, **kwargs: Any) -> httpx.Response:
        return await asyncio.to_thread(self._client.post, url, **kwargs)

    async def aclose(self) -> None:
        await asyncio.to_thread(self._client.close)


class HTTPClientAdapter:
    metadata = AdapterMetadata(
        category="http",
        provider="httpx",
        factory="oneiric.adapters.http.httpx:HTTPClientAdapter",
        capabilities=["http", "rest", "otlp"],
        stack_level=10,
        priority=200,
        source=CandidateSource.LOCAL_PKG,
        owner="Platform Core",
        requires_secrets=False,
        settings_model=HTTPClientSettings,
    )

    def __init__(
        self,
        settings: HTTPClientSettings | None = None,
        *,
        transport: httpx.AsyncBaseTransport | None = None,
    ) -> None:
        self._settings = settings or HTTPClientSettings()
        self._transport = transport
        self._client: Any | None = None
        self._logger = get_logger("adapter.http.httpx").bind(
            domain="adapter",
            key="http",
            provider="httpx",
        )

    async def init(self) -> None:
        """Initialise AsyncClient with optional base URL + defaults."""
        async_client_cls: type[Any] = (
            getattr(httpx, "AsyncClient", None) or _AsyncClientShim
        )
        client_kwargs: dict[str, Any] = {
            "timeout": self._settings.timeout,
            "verify": self._settings.verify,
        }
        if self._settings.base_url:
            client_kwargs["base_url"] = str(self._settings.base_url)
        if self._settings.headers:
            client_kwargs["headers"] = self._settings.headers
        if self._transport:
            client_kwargs["transport"] = self._transport

        self._client = async_client_cls(**client_kwargs)
        self._logger.info(
            "adapter-init",
            adapter="httpx",
            base_url=str(self._settings.base_url or ""),
        )

    async def health(self) -> bool:
        """Run a lightweight GET when base_url is configured."""
        if not self._client:
            return False
        if not self._settings.base_url:
            return True
        try:
            response = await self._client.get(self._settings.healthcheck_path)
            return response.status_code < 500
        except httpx.HTTPError as exc:
            self._logger.warning("adapter-health-failed", error=str(exc))
            return False

    async def cleanup(self) -> None:
        """Dispose AsyncClient (or mocked equivalent)."""
        if self._client is None:
            return
        close_method = getattr(self._client, "aclose", None)
        if callable(close_method):
            maybe_coro = close_method()
            if inspect.isawaitable(maybe_coro):
                await maybe_coro
        elif hasattr(self._client, "close"):
            close_callable = getattr(self._client, "close")
            if callable(close_callable):
                close_callable()
        self._client = None
        self._logger.info("adapter-cleanup-complete", adapter="httpx")

    async def request(self, method: str, url: str, **kwargs: Any) -> httpx.Response:
        client = self._ensure_client()
        caller = getattr(client, "request", None)
        return await self._request(method, url, caller=caller, **kwargs)

    async def get(self, url: str, **kwargs: Any) -> httpx.Response:
        client = self._ensure_client()
        get_fn = getattr(client, "get", None)
        caller = (
            (lambda _method, target, **opts: get_fn(target, **opts))
            if callable(get_fn)
            else None
        )
        return await self._request("GET", url, caller=caller, **kwargs)

    async def post(self, url: str, **kwargs: Any) -> httpx.Response:
        client = self._ensure_client()
        post_fn = getattr(client, "post", None)
        caller = (
            (lambda _method, target, **opts: post_fn(target, **opts))
            if callable(post_fn)
            else None
        )
        return await self._request("POST", url, caller=caller, **kwargs)

    def _ensure_client(self) -> Any:
        if not self._client:
            raise LifecycleError("httpx-client-not-initialized")
        return self._client

    async def _request(
        self,
        method: str,
        url: str,
        *,
        caller: Any | None,
        **kwargs: Any,
    ) -> httpx.Response:
        client = self._ensure_client()
        headers: dict[str, str] = dict(kwargs.pop("headers", {}) or {})
        inject_trace_context(headers)
        if headers:
            kwargs["headers"] = headers
        operation = method.upper()

        async def send() -> httpx.Response:
            if callable(caller):
                return await caller(method, url, **kwargs)
            return await client.request(method, url, **kwargs)

        return await observed_http_request(
            domain="adapter",
            key="http",
            adapter="http",
            provider="httpx",
            operation=operation,
            url=url,
            component="adapter.http",
            span_name="adapter.http.request",
            send=send,
        )



"""Base classes for vector database adapters."""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Any

from pydantic import BaseModel, Field

from oneiric.core.logging import get_logger


class VectorSearchResult(BaseModel):
    """Standard vector search result."""

    id: str
    score: float
    metadata: dict[str, Any] = Field(default_factory=dict)
    vector: list[float] | None = None


class VectorDocument(BaseModel):
    """Standard vector document for insertion."""

    id: str | None = None
    vector: list[float]
    metadata: dict[str, Any] = Field(default_factory=dict)


class VectorBaseSettings(BaseModel):
    """Base settings for vector adapters."""

    collection_prefix: str = ""
    default_dimension: int = 1536  # OpenAI ada-002 default
    default_distance_metric: str = "cosine"  # cosine, euclidean, dot_product

    # Connection settings
    connect_timeout: float = 30.0
    request_timeout: float = 30.0
    max_retries: int = 3

    # Performance settings
    batch_size: int = 100
    max_connections: int = 10


class VectorCollection:
    """Wrapper for vector collection operations."""

    def __init__(self, adapter: Any, name: str) -> None:
        self.adapter = adapter
        self.name = name

    async def search(
        self,
        query_vector: list[float],
        limit: int = 10,
        filter_expr: dict[str, Any] | None = None,
        include_vectors: bool = False,
        **kwargs: Any,
    ) -> list[VectorSearchResult]:
        """Perform vector similarity search."""
        return await self.adapter.search(
            self.name,
            query_vector,
            limit,
            filter_expr,
            include_vectors,
            **kwargs,
        )

    async def insert(
        self,
        documents: list[VectorDocument],
        **kwargs: Any,
    ) -> list[str]:
        """Insert documents with vectors."""
        return await self.adapter.insert(
            self.name,
            documents,
            **kwargs,
        )

    async def upsert(
        self,
        documents: list[VectorDocument],
        **kwargs: Any,
    ) -> list[str]:
        """Upsert documents with vectors."""
        return await self.adapter.upsert(
            self.name,
            documents,
            **kwargs,
        )

    async def delete(self, ids: list[str], **kwargs: Any) -> bool:
        """Delete documents by IDs."""
        return await self.adapter.delete(
            self.name,
            ids,
            **kwargs,
        )

    async def get(
        self,
        ids: list[str],
        include_vectors: bool = False,
        **kwargs: Any,
    ) -> list[VectorDocument]:
        """Retrieve documents by IDs."""
        return await self.adapter.get(
            self.name,
            ids,
            include_vectors,
            **kwargs,
        )

    async def count(
        self,
        filter_expr: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> int:
        """Count documents in collection."""
        return await self.adapter.count(
            self.name,
            filter_expr,
            **kwargs,
        )


class VectorBase:
    """Base class for vector database adapters with lifecycle integration."""

    def __init__(self, settings: VectorBaseSettings) -> None:
        self._settings = settings
        self._collections: dict[str, VectorCollection] = {}
        self._client: Any | None = None
        self._logger = get_logger("adapter.vector.base")

    def __getattr__(self, name: str) -> Any:
        """Dynamic collection access."""
        if name not in self._collections:
            self._collections[name] = VectorCollection(self, name)
        return self._collections[name]

    async def get_client(self) -> Any:
        """Get the underlying vector database client."""
        return await self._ensure_client()

    # Lifecycle hooks (to be implemented by Oneiric lifecycle manager)
    @abstractmethod
    async def init(self) -> None:
        """Initialize the vector adapter."""
        ...

    @abstractmethod
    async def health(self) -> bool:
        """Check if the vector database is healthy."""
        ...

    @abstractmethod
    async def cleanup(self) -> None:
        """Cleanup vector database resources."""
        ...

    # Core vector operations (to be implemented by concrete adapters)
    @abstractmethod
    async def search(
        self,
        collection: str,
        query_vector: list[float],
        limit: int = 10,
        filter_expr: dict[str, Any] | None = None,
        include_vectors: bool = False,
        **kwargs: Any,
    ) -> list[VectorSearchResult]:
        """Perform vector similarity search."""
        ...

    @abstractmethod
    async def insert(
        self,
        collection: str,
        documents: list[VectorDocument],
        **kwargs: Any,
    ) -> list[str]:
        """Insert documents with vectors."""
        ...

    @abstractmethod
    async def upsert(
        self,
        collection: str,
        documents: list[VectorDocument],
        **kwargs: Any,
    ) -> list[str]:
        """Upsert documents with vectors."""
        ...

    @abstractmethod
    async def delete(
        self,
        collection: str,
        ids: list[str],
        **kwargs: Any,
    ) -> bool:
        """Delete documents by IDs."""
        ...

    @abstractmethod
    async def get(
        self,
        collection: str,
        ids: list[str],
        include_vectors: bool = False,
        **kwargs: Any,
    ) -> list[VectorDocument]:
        """Retrieve documents by IDs."""
        ...

    @abstractmethod
    async def count(
        self,
        collection: str,
        filter_expr: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> int:
        """Count documents in collection."""
        ...

    @abstractmethod
    async def create_collection(
        self,
        name: str,
        dimension: int,
        distance_metric: str = "cosine",
        **kwargs: Any,
    ) -> bool:
        """Create a new collection."""
        ...

    @abstractmethod
    async def delete_collection(
        self,
        name: str,
        **kwargs: Any,
    ) -> bool:
        """Delete a collection."""
        ...

    @abstractmethod
    async def list_collections(self, **kwargs: Any) -> list[str]:
        """List all collections."""
        ...

    @abstractmethod
    async def _ensure_client(self) -> Any:
        """Ensure client is initialized."""
        ...

    @abstractmethod
    async def _create_client(self) -> Any:
        """Create the underlying client."""
        ...

    def has_capability(self, capability: str) -> bool:
        """Check if adapter supports a specific capability."""
        return False  # Base implementation - override in adapters

    @asynccontextmanager
    async def transaction(self) -> AsyncGenerator[Any]:
        """Transaction context manager (if supported)."""
        client = await self.get_client()
        yield client



"""SQLite database adapter."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any

from pydantic import BaseModel, Field

from oneiric.adapters.metadata import AdapterMetadata
from oneiric.core.lifecycle import LifecycleError
from oneiric.core.logging import get_logger
from oneiric.core.resolution import CandidateSource


class SQLiteDatabaseSettings(BaseModel):
    """Configuration for the SQLite adapter."""

    path: str = Field(
        default=":memory:", description="Path to the SQLite database file."
    )
    isolation_level: str | None = Field(default=None)
    pragmas: dict[str, str | int | float] = Field(default_factory=dict)


class SQLiteDatabaseAdapter:
    """A lightweight SQLite adapter for dev/test scenarios."""

    metadata = AdapterMetadata(
        category="database",
        provider="sqlite",
        factory="oneiric.adapters.database.sqlite:SQLiteDatabaseAdapter",
        capabilities=["sql"],
        stack_level=5,
        priority=50,
        source=CandidateSource.LOCAL_PKG,
        owner="Data Platform",
        requires_secrets=False,
        settings_model=SQLiteDatabaseSettings,
    )

    def __init__(
        self,
        settings: SQLiteDatabaseSettings | None = None,
        *,
        connection_factory: Callable[..., Awaitable[Any]] | None = None,
    ) -> None:
        self._settings = settings or SQLiteDatabaseSettings()
        self._connection_factory = connection_factory
        self._conn: Any = None
        self._logger = get_logger("adapter.database.sqlite").bind(
            domain="adapter",
            key="database",
            provider="sqlite",
        )

    async def init(self) -> None:
        if self._conn:
            return
        factory = self._connection_factory
        if factory is None:
            try:
                import aiosqlite
            except ModuleNotFoundError as exc:  # pragma: no cover - defensive
                raise LifecycleError("aiosqlite-missing") from exc
            factory = aiosqlite.connect
        self._conn = await factory(
            self._settings.path, isolation_level=self._settings.isolation_level
        )
        for pragma, value in self._settings.pragmas.items():
            await self._conn.execute(f"PRAGMA {pragma}={value}")
        await self._conn.commit()
        self._logger.info(
            "adapter-init", adapter="sqlite-database", path=self._settings.path
        )

    async def health(self) -> bool:
        conn = self._ensure_conn()
        try:
            cursor = await conn.execute("SELECT 1;")
            await cursor.close()
            return True
        except Exception as exc:  # pragma: no cover - disk errors
            self._logger.warning("adapter-health-error", error=str(exc))
            return False

    async def cleanup(self) -> None:
        if not self._conn:
            return
        await self._conn.close()
        self._conn = None
        self._logger.info("adapter-cleanup-complete", adapter="sqlite-database")

    async def execute(self, query: str, *args: Any) -> int:
        conn = self._ensure_conn()
        cursor = await conn.execute(query, args)
        await conn.commit()
        rowcount = cursor.rowcount if cursor.rowcount is not None else 0
        await cursor.close()
        return rowcount

    async def fetch_all(self, query: str, *args: Any) -> list[Any]:
        conn = self._ensure_conn()
        cursor = await conn.execute(query, args)
        rows = await cursor.fetchall()
        await cursor.close()
        return rows

    async def fetch_one(self, query: str, *args: Any) -> Any:
        conn = self._ensure_conn()
        cursor = await conn.execute(query, args)
        row = await cursor.fetchone()
        await cursor.close()
        return row

    def _ensure_conn(self) -> Any:
        if not self._conn:
            raise LifecycleError("sqlite-connection-not-initialized")
        return self._conn



"""Shared persistence helpers for domain pause/drain state."""

from __future__ import annotations

import sqlite3
import threading
from collections.abc import Iterator, Mapping
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any


@dataclass
class DomainActivity:
    paused: bool = False
    draining: bool = False
    note: str | None = None

    def as_dict(self) -> dict[str, Any]:
        return {
            "paused": self.paused,
            "draining": self.draining,
            "note": self.note,
        }

    @classmethod
    def from_mapping(cls, data: Mapping[str, Any]) -> DomainActivity:
        return cls(
            paused=bool(data.get("paused", False)),
            draining=bool(data.get("draining", False)),
            note=data.get("note"),
        )

    def is_default(self) -> bool:
        return not self.paused and not self.draining and self.note is None


class DomainActivityStore:
    """SQLite-backed persistence for domain pause/drain activity."""

    def __init__(self, path: str | Path) -> None:
        self.path = Path(path)
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self._lock = threading.RLock()
        self._ensure_schema()

    def get(self, domain: str, key: str) -> DomainActivity:
        with self._connection() as conn:
            conn: sqlite3.Connection
            row = conn.execute(
                "SELECT paused, draining, note FROM activity WHERE domain=? AND key=?",
                (domain, key),
            ).fetchone()
        if row:
            return DomainActivity(
                paused=bool(row[0]),
                draining=bool(row[1]),
                note=row[2],
            )
        return DomainActivity()

    def all_for_domain(self, domain: str) -> dict[str, DomainActivity]:
        with self._connection() as conn:
            conn: sqlite3.Connection
            rows = conn.execute(
                "SELECT key, paused, draining, note FROM activity WHERE domain=?",
                (domain,),
            ).fetchall()
        return {
            row[0]: DomainActivity(
                paused=bool(row[1]), draining=bool(row[2]), note=row[3]
            )
            for row in rows
        }

    def set(self, domain: str, key: str, state: DomainActivity) -> None:
        with self._connection() as conn:
            conn: sqlite3.Connection
            if state.is_default():
                conn.execute(
                    "DELETE FROM activity WHERE domain=? AND key=?", (domain, key)
                )
                conn.commit()
                return
            conn.execute(
                """
                INSERT INTO activity(domain, key, paused, draining, note)
                VALUES (?, ?, ?, ?, ?)
                ON CONFLICT(domain, key) DO UPDATE SET
                    paused=excluded.paused,
                    draining=excluded.draining,
                    note=excluded.note
                """,
                (domain, key, int(state.paused), int(state.draining), state.note),
            )
            conn.commit()

    def snapshot(self) -> dict[str, dict[str, DomainActivity]]:
        with self._connection() as conn:
            conn: sqlite3.Connection
            rows = conn.execute(
                "SELECT domain, key, paused, draining, note FROM activity",
            ).fetchall()
        snapshot: dict[str, dict[str, DomainActivity]] = {}
        for domain, key, paused, draining, note in rows:
            domain_map = snapshot.setdefault(domain, {})
            domain_map[key] = DomainActivity(
                paused=bool(paused),
                draining=bool(draining),
                note=note,
            )
        return snapshot

    # internal -----------------------------------------------------------------

    @contextmanager
    def _connection(self) -> Iterator[sqlite3.Connection]:
        with self._lock:
            conn = sqlite3.connect(self.path)
            conn.row_factory = sqlite3.Row
            try:
                yield conn
            finally:
                conn.close()

    def _ensure_schema(self) -> None:
        with self._connection() as conn:
            conn: sqlite3.Connection
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS activity (
                    domain TEXT NOT NULL,
                    key TEXT NOT NULL,
                    paused INTEGER NOT NULL DEFAULT 0,
                    draining INTEGER NOT NULL DEFAULT 0,
                    note TEXT,
                    PRIMARY KEY (domain, key)
                )
                """
            )
            conn.commit()
